"""
Module containing tests for the DatabaseBaseModel class in the model module
in the database destination package
"""
import os
from datetime import date
from typing import Dict, Type, Optional, List
from unittest import TestCase, main

import sqlalchemy as orm
from sqlalchemy.ext.declarative import declarative_base
from dotenv import load_dotenv

from judah.destinations.database.connection import DatabaseConnection
from judah.destinations.database.config import DatabaseConnectionConfig
from judah.destinations.database.model import DatabaseBaseModel
from test.utils import create_tables_in_database, delete_tables_from_database, drop_schema

_ROOT_PATH = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

load_dotenv(os.path.join(_ROOT_PATH, '.env'))

_TEST_DB_URI = os.environ.get('TEST_POSTGRES_DB_URI')

_TEST_DB_BASE = declarative_base()
_TABLE_NAME = 'dummy'
_SCHEMA_NAME = 'test_schema'


class TestModel(DatabaseBaseModel, _TEST_DB_BASE):
    """Test database model"""
    __tablename__ = _TABLE_NAME
    __table_args__: Dict = {'schema': _SCHEMA_NAME}
    _db_configuration: DatabaseConnectionConfig = DatabaseConnectionConfig(db_uri=_TEST_DB_URI)
    _base_declarative_class: Type[declarative_base()] = _TEST_DB_BASE
    _datetime_fields: Optional[List[str]] = ["Date"]

    Date = orm.Column(orm.Date, primary_key=True)
    number = orm.Column(orm.Integer, primary_key=True)
    Capacity = orm.Column(orm.Integer)
    Price = orm.Column(orm.Integer)


class TestDatabaseBaseModel(TestCase):
    """Tests for the DatabaseBaseModel base class"""

    def setUp(self) -> None:
        """Initialize a few variables"""
        self.data = [
            {"Date": date(year=2020, month=3, day=9), "number": 1, "Capacity": 16616, "Price": 67},
            {"Date": date(year=2020, month=3, day=12), "number": 2, "Capacity": 16516, "Price": 567},
            {"Date": date(year=2020, month=3, day=10), "number": 3, "Capacity": 16616, "Price": 637},
            {"Date": date(year=2020, month=3, day=9), "number": 4, "Capacity": 16620, "Price": 617},
        ]
        try:
            delete_tables_from_database(db_configuration=TestModel._db_configuration,
                                        table_name=_TABLE_NAME, schema_name=_SCHEMA_NAME)
        except Exception:
            pass

    def load_database(self):
        """Loads the database with the self.data"""
        create_tables_in_database(db_configuration=TestModel._db_configuration,
                                  model_base=_TEST_DB_BASE, schema_name=_SCHEMA_NAME)

        with DatabaseConnection.get_db_connection(
                db_connection_config=TestModel._db_configuration) as db_connection:
            for datum in self.data:
                record = TestModel(**datum)
                db_connection.db_session.add(record)

            db_connection.db_session.commit()

    def test_get_attributes(self):
        """Should return the column names of the model"""
        self.load_database()

        column_names = TestModel.get_attributes()
        expected_columns = ['Date', 'number', 'Capacity', 'Price', 'created_at', 'updated_at']
        self.assertListEqual(sorted(column_names), sorted(expected_columns))

    def test_get_last_record(self):
        """Should return the latest record according to the given datetime column"""
        self.load_database()

        last_record = TestModel.get_last_record()
        expected_last_record = self.data[1]

        columns = ['Date', 'number', 'Capacity', 'Price']
        for column in columns:
            self.assertEqual(getattr(last_record, column), expected_last_record[column])

    def test_get_last_saved_timestamp(self):
        """Should return the timestamp of the last saved record"""
        self.load_database()

        last_timestamp = TestModel.get_last_saved_timestamp()
        expected_last_timestamp = self.data[1]['Date']

        self.assertEqual(last_timestamp, expected_last_timestamp)

    def test_update(self):
        """Should update the attributes passed in the kwargs and saves"""
        create_tables_in_database(db_configuration=TestModel._db_configuration,
                                  model_base=_TEST_DB_BASE, schema_name=_SCHEMA_NAME)

        with DatabaseConnection.get_db_connection(
                db_connection_config=TestModel._db_configuration) as db_connection:
            record = TestModel(**self.data[1])
            db_connection.db_session.add(record)
            db_connection.db_session.commit()

            new_capacity = 56
            new_price = 7

            record.update(session=db_connection.db_session, Capacity=new_capacity, Price=new_price)

            record_from_database = db_connection.db_session.query(TestModel).filter_by(
                Date=record.Date, number=record.number).first()
            db_connection.db_session.commit()

            self.assertEqual(record_from_database.Capacity, new_capacity)
            self.assertEqual(record_from_database.Price, new_price)

    def test_save(self):
        """Should commit any new changes to the database"""
        create_tables_in_database(db_configuration=TestModel._db_configuration,
                                  model_base=_TEST_DB_BASE, schema_name=_SCHEMA_NAME)

        with DatabaseConnection.get_db_connection(
                db_connection_config=TestModel._db_configuration) as db_connection:
            record = TestModel(**self.data[1])

            record_from_database_pre_save = db_connection.db_session.query(TestModel).filter_by(
                Date=record.Date, number=record.number).first()
            db_connection.db_session.commit()

            record.save(db_connection.db_session)

            record_from_database_post_save = db_connection.db_session.query(TestModel).filter_by(
                Date=record.Date, number=record.number).first()
            db_connection.db_session.commit()

            self.assertIsNone(record_from_database_pre_save)
            self.assertIsInstance(record_from_database_post_save, TestModel)

    def test_delete(self):
        """Should delete the current record from the database"""
        create_tables_in_database(db_configuration=TestModel._db_configuration,
                                  model_base=_TEST_DB_BASE, schema_name=_SCHEMA_NAME)

        with DatabaseConnection.get_db_connection(
                db_connection_config=TestModel._db_configuration) as db_connection:
            record = TestModel(**self.data[1])
            db_connection.db_session.add(record)
            db_connection.db_session.commit()

            record_from_database_pre_deletion = db_connection.db_session.query(TestModel).filter_by(
                Date=record.Date, number=record.number).first()

            record.delete(db_connection.db_session)

            record_from_database_post_deletion = db_connection.db_session.query(TestModel).filter_by(
                Date=record.Date, number=record.number).first()

            self.assertIsNone(record_from_database_post_deletion)
            self.assertIsInstance(record_from_database_pre_deletion, TestModel)

    def test_create_schema(self):
        """Should create the schema for this class if it exists"""
        drop_schema(db_configuration=TestModel._db_configuration, schema_name=_SCHEMA_NAME)

        with DatabaseConnection.get_db_connection(
                db_connection_config=TestModel._db_configuration) as db_connection:
            schema_check_sql = f"""
            SELECT exists(select schema_name FROM information_schema.schemata WHERE schema_name = '{_SCHEMA_NAME}')
            """
            self.assertFalse(db_connection.execute_sql(schema_check_sql).first()[0])

            TestModel.create_schema(db_connection.connection_engine)

            self.assertTrue(db_connection.execute_sql(schema_check_sql).first()[0])

    def test_initialize(self):
        """Should create the tables in the database"""
        drop_schema(db_configuration=TestModel._db_configuration, schema_name=_SCHEMA_NAME)

        with DatabaseConnection.get_db_connection(
                db_connection_config=TestModel._db_configuration) as db_connection:
            table_check_sql = f"""
            SELECT EXISTS (
               SELECT FROM pg_tables
               WHERE  schemaname = '{_SCHEMA_NAME}'
               AND    tablename  = '{_TABLE_NAME}'
               )
            """
            self.assertFalse(db_connection.execute_sql(table_check_sql).first()[0])

            TestModel.initialize()

            self.assertTrue(db_connection.execute_sql(table_check_sql).first()[0])

    def test_upsert_new_record(self):
        """upsert should creates a new record if it does not exist and then return the data"""
        create_tables_in_database(db_configuration=TestModel._db_configuration,
                                  model_base=_TEST_DB_BASE, schema_name=_SCHEMA_NAME)

        with DatabaseConnection.get_db_connection(
                db_connection_config=TestModel._db_configuration) as db_connection:
            raw_data = self.data[1]
            record_from_database_pre_insert = db_connection.db_session.query(TestModel).filter_by(
                Date=raw_data['Date'], number=raw_data['number']).first()
            db_connection.db_session.commit()
            self.assertIsNone(record_from_database_pre_insert)

            recorded_data = TestModel.upsert(raw_data)
            record_from_database_post_insert = db_connection.db_session.query(TestModel).filter_by(
                Date=raw_data['Date'], number=raw_data['number']).first()
            db_connection.db_session.commit()
            self.assertDictEqual(recorded_data, raw_data)
            self.assertIsInstance(record_from_database_post_insert, TestModel)

            for field, value in raw_data.items():
                self.assertEqual(getattr(record_from_database_post_insert, field), value)

    def test_upsert_old_record(self):
        """upsert should update an existing record if it does not exist and then return the data"""
        create_tables_in_database(db_configuration=TestModel._db_configuration,
                                  model_base=_TEST_DB_BASE, schema_name=_SCHEMA_NAME)

        with DatabaseConnection.get_db_connection(
                db_connection_config=TestModel._db_configuration) as db_connection:
            raw_data = self.data[1]
            record = TestModel(**raw_data)
            db_connection.db_session.add(record)
            db_connection.db_session.commit()

            new_data = {
                **raw_data,
                'Capacity': 7643,
                'Price': 211
            }
            updated_data = TestModel.upsert(new_data)
            record_from_database_post_update = db_connection.db_session.query(TestModel).filter_by(
                Date=raw_data['Date'], number=raw_data['number']).first()
            self.assertDictEqual(updated_data, new_data)
            self.assertIsInstance(record_from_database_post_update, TestModel)

            for field, value in new_data.items():
                self.assertEqual(getattr(record_from_database_post_update, field), value)

    def tearDown(self) -> None:
        try:
            delete_tables_from_database(db_configuration=TestModel._db_configuration,
                                        table_name=_TABLE_NAME, schema_name=_SCHEMA_NAME)
        except Exception:
            pass

        DatabaseConnection.close_all_connections()
        DatabaseConnection.remove_all_connections()


if __name__ == '__main__':
    main()



from datetime import date
from typing import Optional, Iterator, Dict, Any

from ..base import BaseSource
from ...utils.dates import update_date


class DateBasedBaseSource(BaseSource):
    start_date: Optional[date] = None
    end_date: Optional[date] = None
    default_batch_size_in_days: int = 7

    def get(self) -> Iterator[Dict[str, Any]]:
        """Returns data from a given date to a given date as an iterator"""
        start_date = self._get_next_start_date()
        end_date = self._get_next_end_date()

        if end_date < start_date:
            yield from []
        else:
            yield from self._query_data_source(start_date=start_date, end_date=end_date)

            self.start_date = self._get_next_start_date(days_to_increment_by=self.default_batch_size_in_days)

    def _query_data_source(self, start_date: date, end_date: date) -> Iterator[Dict[str, Any]]:
        """Queries a given start and end date and returns an iterator with data records"""
        raise NotImplementedError('Implement the _query_data_source method to return an iterator of dictionaries')

    def _get_next_end_date(self, days_to_increment_by: int = 0, days_to_decrement_by: int = 0) -> date:
        """Gets the next end date in the given iteration"""
        initial_date: Optional[date] = self.end_date

        if not isinstance(initial_date, date):
            if isinstance(self.start_date, date):
                initial_date = self.start_date
                days_to_increment_by = self.default_batch_size_in_days + days_to_increment_by
            else:
                # Assumption: when no start or end dates are given, the range is start -> today
                initial_date = date.today()

        return update_date(date_to_update=initial_date, days_to_decrement_by=days_to_decrement_by,
                           days_to_increment_by=days_to_increment_by)

    def _get_next_start_date(self, days_to_increment_by: int = 0, days_to_decrement_by: int = 0) -> date:
        """Gets the next start date in the given iteration"""
        initial_date: Optional[date] = self.start_date

        if not isinstance(initial_date, date):
            # assumption: today is the end_date
            initial_date = self.end_date if isinstance(self.end_date, date) else date.today()
            days_to_decrement_by = self.default_batch_size_in_days + days_to_decrement_by

        return update_date(date_to_update=initial_date, days_to_decrement_by=days_to_decrement_by,
                           days_to_increment_by=days_to_increment_by)



"""Module containing tests for the date based controller"""
import collections
import os
from datetime import date, datetime, timedelta
from typing import Optional, Type, Iterator, Dict, Any, List, Union
from unittest import TestCase, main
from unittest.mock import patch, Mock

import pytz
from dotenv import load_dotenv
from freezegun import freeze_time
from sqlalchemy.ext.declarative import declarative_base
import sqlalchemy as orm

from judah.controllers.base.date_based import DateBasedBaseController
from judah.sources.base.date_based import DateBasedBaseSource
from judah.destinations.database.model import DatabaseBaseModel
from judah.destinations.database.connection import DatabaseConnectionConfig, DatabaseConnection
from judah.transformers.base import BaseTransformer
from judah.utils.cron import CronType, Hour
from test.utils import delete_tables_from_database, create_tables_in_database, drop_schema

_ROOT_PATH = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))

load_dotenv(os.path.join(_ROOT_PATH, '.env'))

_TEST_DB_URI = os.environ.get('TEST_POSTGRES_DB_URI')

_TEST_DB_BASE = declarative_base()
_TABLE_NAME = 'dummy2'
_SCHEMA_NAME = 'test_schema'


class ChildDateBasedModel(DatabaseBaseModel, _TEST_DB_BASE):
    """Child Date Based Model"""
    __tablename__ = _TABLE_NAME
    __table_args__: Dict = {'schema': _SCHEMA_NAME}
    _db_configuration: DatabaseConnectionConfig = DatabaseConnectionConfig(db_uri=_TEST_DB_URI)
    _base_declarative_class: Type[declarative_base()] = _TEST_DB_BASE
    _datetime_fields: Optional[List[str]] = ["Date"]

    Date = orm.Column(orm.Date, primary_key=True)
    number = orm.Column(orm.Integer, primary_key=True)
    Capacity = orm.Column(orm.Integer)
    Price = orm.Column(orm.Integer)


class ChildDateBasedSource(DateBasedBaseSource):
    """Child Date based source"""
    start_date: date = date(year=2019, month=8, day=9)
    end_date: date = date(year=2019, month=8, day=19)

    def _query_data_source(self, start_date: date, end_date: date) -> Iterator[Dict[str, Any]]:
        pass


class ChildDateBasedController(DateBasedBaseController):
    """Child date based controller"""
    start_date: Optional[date] = date(year=2020, month=1, day=1)
    end_date: Optional[date] = date(year=2020, month=1, day=1)
    source_class: Type[DateBasedBaseSource] = ChildDateBasedSource
    destination_model_class = ChildDateBasedModel


class TestDateBasedController(TestCase):
    """Tests for the DateBasedController"""

    def setUp(self) -> None:
        """Initialize some variables"""
        self.data_from_source = [
            {"Date": date(year=2020, month=3, day=9), "number": 1, "Capacity": 16616, "Price": 67},
            {"Date": date(year=2020, month=3, day=12), "number": 2, "Capacity": 16516, "Price": 567},
            {"Date": date(year=2020, month=3, day=10), "number": 3, "Capacity": 16616, "Price": 637},
            {"Date": date(year=2020, month=3, day=9), "number": 4, "Capacity": 16620, "Price": 617},
        ]

    @patch.object(ChildDateBasedModel, 'get_last_saved_timestamp')
    def test_update_last_timestamp(self, mock_get_last_saved_timestamp):
        """Should update the last timestamp property of the Controller class and the start_date"""
        timestamp = datetime.now()
        mock_get_last_saved_timestamp.return_value = timestamp

        self.assertIsNone(ChildDateBasedController._last_timestamp)
        self.assertNotEqual(ChildDateBasedController.start_date, timestamp)

        ChildDateBasedController._update_last_timestamp()
        mock_get_last_saved_timestamp.assert_called_once()

        self.assertEqual(ChildDateBasedController._last_timestamp, timestamp)
        self.assertEqual(ChildDateBasedController.start_date, timestamp)

    @patch.object(ChildDateBasedController, '_update_source')
    @patch.object(ChildDateBasedSource, '_query_data_source')
    def test_query_source(self, mock_query_data_source, mock_update_source):
        """
        Should query the data source for the given dates and return the iterator of data dictionaries
        as well as increment the start date by the default batch size in days
        """
        source = ChildDateBasedSource()
        mock_query_data_source.return_value = (record for record in self.data_from_source)
        # set the source of the controller
        ChildDateBasedController._source = source
        original_start_date = source.start_date
        original_end_date = source.end_date

        response = ChildDateBasedController._query_source()
        self.assertIsInstance(response, collections.Iterator)

        received_data = [record for record in response]

        mock_query_data_source.assert_called_once_with(
            start_date=original_start_date, end_date=original_end_date)
        mock_update_source.assert_called_once_with(
            start_date=ChildDateBasedController.start_date, end_date=ChildDateBasedController.end_date)

        self.assertListEqual(received_data, self.data_from_source)
        self.assertEqual(
            source.start_date, original_start_date + timedelta(days=source.default_batch_size_in_days))

    @patch.object(ChildDateBasedSource, '_query_data_source')
    def test_extract(self, mock_query_data_source):
        """Should extract data from the source and return the iterator of data"""
        mock_query_data_source.return_value = (record for record in self.data_from_source)

        response = ChildDateBasedController.extract()
        self.assertIsInstance(response, collections.Iterator)

        received_data = [record for record in response]

        self.assertListEqual(received_data, self.data_from_source)
        mock_query_data_source.assert_called_once_with(
            start_date=ChildDateBasedController.start_date, end_date=ChildDateBasedController.end_date)

    def test_transform(self):
        """Should transform the data row to another form"""
        sample_data = self.data_from_source[0]
        expected_data = {"Date": date(year=2020, month=3, day=9), "number": "1.0", "Capacity": "16616.0", "Price": "67.0"}

        class IntegerToFloatTransformer(BaseTransformer):
            """Transforms all integer values to floats"""

            @classmethod
            def run(cls, data: Dict[Any, Any]) -> Union[Dict[Any, Any], List[Any]]:
                return {key: float(value) if isinstance(value, int) else value for key, value in data.items()}

        class FloatToStringTransformer(BaseTransformer):
            """Transforms all float values to strings"""

            @classmethod
            def run(cls, data: Dict[Any, Any]) -> Union[Dict[Any, Any], List[Any]]:
                return {key: str(value) if isinstance(value, float) else value for key, value in data.items()}

        ChildDateBasedController.transformer_classes = [IntegerToFloatTransformer, FloatToStringTransformer]
        response = ChildDateBasedController.transform(sample_data)
        self.assertIsInstance(response, collections.Iterator)

        received_data = [record for record in response]
        self.assertEqual(len(received_data), 1)
        self.assertDictEqual(received_data[0], expected_data)

    def test_load(self):
        """Should load data into the destination and return the data iterator that was put in the destination"""
        create_tables_in_database(db_configuration=ChildDateBasedModel._db_configuration,
                                  model_base=_TEST_DB_BASE, schema_name=_SCHEMA_NAME)

        with DatabaseConnection.get_db_connection(
                db_connection_config=ChildDateBasedModel._db_configuration) as db_connection:
            raw_data = self.data_from_source[0]

            record_from_database_pre_load = db_connection.db_session.query(ChildDateBasedModel).filter_by(
                Date=raw_data['Date'], number=raw_data['number']).first()
            db_connection.db_session.commit()
            self.assertIsNone(record_from_database_pre_load)

            response = ChildDateBasedController.load(raw_data)

            loaded_data = [record for record in response]
            self.assertEqual(len(loaded_data), 1)
            self.assertDictEqual(loaded_data[0], raw_data)

            record_from_database_post_load = db_connection.db_session.query(ChildDateBasedModel).filter_by(
                Date=raw_data['Date'], number=raw_data['number']).first()
            db_connection.db_session.commit()
            self.assertIsInstance(record_from_database_post_load, ChildDateBasedModel)

            for field, value in raw_data.items():
                self.assertEqual(getattr(record_from_database_post_load, field), value)

    def test_update_source(self):
        """
        Should update the configuration of the source especially
        those attributes that are bound to change and update the source's attributes with those from the model
        """
        self.assertRaises(AttributeError, getattr, ChildDateBasedController, '_source')
        start_date = date(year=1990, month=4, day=7)
        end_date = date(year=2000, month=5, day=8)
        default_batch_size_in_days = 3
        expected_attributes = ChildDateBasedModel.get_attributes()

        ChildDateBasedController._update_source(start_date=start_date, end_date=end_date,
                                                default_batch_size_in_days=default_batch_size_in_days)

        source = ChildDateBasedController._source

        self.assertIsInstance(source, ChildDateBasedSource)
        self.assertEqual(source.start_date, start_date)
        self.assertEqual(source.end_date, end_date)
        self.assertEqual(source.default_batch_size_in_days, default_batch_size_in_days)
        self.assertListEqual(source.attributes, expected_attributes)

    def test_update_interval(self):
        """Should update the _interval basing on the interval_in_milliseconds"""
        ChildDateBasedController.interval_in_milliseconds = 3589

        self.assertIsNone(ChildDateBasedController._interval)

        ChildDateBasedController._update_interval()

        self.assertEqual(ChildDateBasedController._interval,
                         timedelta(milliseconds=ChildDateBasedController.interval_in_milliseconds))

    @patch.object(ChildDateBasedModel, 'get_last_saved_timestamp')
    def test_initialize(self, mock_get_last_saved_timestamp):
        """
        Should create the tables if these are non-existent and
        update the _interval and _last_timestamp if necessary
        """
        drop_schema(db_configuration=ChildDateBasedModel._db_configuration, schema_name=_SCHEMA_NAME)

        timestamp = datetime.now()
        mock_get_last_saved_timestamp.return_value = timestamp
        ChildDateBasedController.interval_in_milliseconds = 3589

        self.assertIsNone(ChildDateBasedController._interval)
        self.assertIsNone(ChildDateBasedController._last_timestamp)

        with DatabaseConnection.get_db_connection(
                db_connection_config=ChildDateBasedModel._db_configuration) as db_connection:
            table_check_sql = f"""
            SELECT EXISTS (
               SELECT FROM pg_tables
               WHERE  schemaname = '{_SCHEMA_NAME}'
               AND    tablename  = '{_TABLE_NAME}'
               )
            """
            self.assertFalse(db_connection.execute_sql(table_check_sql).first()[0])

            ChildDateBasedController._initialize()

            self.assertTrue(db_connection.execute_sql(table_check_sql).first()[0])

        mock_get_last_saved_timestamp.assert_called_once()
        self.assertEqual(ChildDateBasedController._interval,
                         timedelta(milliseconds=ChildDateBasedController.interval_in_milliseconds))
        self.assertEqual(ChildDateBasedController._last_timestamp, timestamp)

    @freeze_time('1997-04-13 08:46:30+00:00')
    def test_get_closest_next_timestamp(self):
        """
        Should get the timestamp from the available CronTypes that is the next to be run
        """
        utc_now = datetime.utcnow().astimezone(pytz.utc)
        cron_job_for_next_second = CronType(second=utc_now.second + 1)
        cron_job_after_8_second = CronType(second=utc_now.second + 8)
        cron_job_for_next_day = CronType(hour=Hour(utc_now.hour) - 2)

        ChildDateBasedController.cron_jobs = [
            cron_job_after_8_second,
            cron_job_for_next_second,
            cron_job_for_next_day,
        ]

        self.assertEqual(
            ChildDateBasedController._get_closest_next_timestamp(),
            cron_job_for_next_second.get_next_utc_timestamp())

    def test_wait_for_interval_to_elapse(self):
        """
        Should wait for the interval to elapse before starting again, given a start time.
        margin of error is 1 second
        """
        time_in_seconds_to_wait = 2
        expected_timedelta_to_wait = timedelta(seconds=time_in_seconds_to_wait)
        ChildDateBasedController._interval = expected_timedelta_to_wait

        timestamp_before_waiting = datetime.now(tz=pytz.utc)
        ChildDateBasedController._wait_for_interval_to_elapse(start_time_in_utc=timestamp_before_waiting)
        timestamp_after_waiting = datetime.now(tz=pytz.utc)

        actual_time_elapsed = timestamp_after_waiting - timestamp_before_waiting

        self.assertGreaterEqual(actual_time_elapsed, expected_timedelta_to_wait)
        self.assertLessEqual(actual_time_elapsed, expected_timedelta_to_wait + timedelta(seconds=1))

    @patch.object(ChildDateBasedController, '_get_closest_next_timestamp')
    def test_wait_for_interval_to_elapse_for_cron_jobs(self, mock_get_closest_next_timestamp):
        """
        Should wait for the interval between start time and the cron_job's next timestamp
        to elapse before starting again.
        margin of error is 1 second
        """
        time_in_seconds_to_wait = 2
        expected_timedelta_to_wait = timedelta(seconds=time_in_seconds_to_wait)

        timestamp_before_waiting = datetime.now(tz=pytz.utc)
        mock_get_closest_next_timestamp.return_value = (
                timestamp_before_waiting.astimezone(pytz.utc) + expected_timedelta_to_wait)

        ChildDateBasedController._wait_for_interval_to_elapse(start_time_in_utc=timestamp_before_waiting)
        timestamp_after_waiting = datetime.now(tz=pytz.utc)

        actual_time_elapsed = timestamp_after_waiting - timestamp_before_waiting

        self.assertGreaterEqual(actual_time_elapsed, expected_timedelta_to_wait)
        self.assertLessEqual(actual_time_elapsed, expected_timedelta_to_wait + timedelta(seconds=1))

    @patch.object(ChildDateBasedController, '_get_closest_next_timestamp')
    def test_wait_for_interval_to_elapse_for_both_cron_jobs_and_interval(self, mock_get_closest_next_timestamp):
        """
        Should wait only for the interval generated by the cron_job not that from the interval property
        margin of error is 1 second
        """
        time_in_seconds_to_wait = 2
        expected_timedelta_to_wait = timedelta(seconds=time_in_seconds_to_wait)
        ChildDateBasedController._interval = timedelta(seconds=10)

        timestamp_before_waiting = datetime.now(tz=pytz.utc)
        mock_get_closest_next_timestamp.return_value = (
                timestamp_before_waiting.astimezone(pytz.utc) + expected_timedelta_to_wait)

        ChildDateBasedController._wait_for_interval_to_elapse(start_time_in_utc=timestamp_before_waiting)
        timestamp_after_waiting = datetime.now(tz=pytz.utc)

        actual_time_elapsed = timestamp_after_waiting - timestamp_before_waiting

        self.assertGreaterEqual(actual_time_elapsed, expected_timedelta_to_wait)
        self.assertLess(actual_time_elapsed, ChildDateBasedController._interval)
        self.assertLessEqual(actual_time_elapsed, expected_timedelta_to_wait + timedelta(seconds=1))

    def tearDown(self) -> None:
        # reset the Controller class
        ChildDateBasedController.transformer_classes = []
        ChildDateBasedController._interval = None
        ChildDateBasedController._last_timestamp = None
        ChildDateBasedController.cron_jobs = []

        try:
            del ChildDateBasedController._source
        except AttributeError:
            pass

        try:
            delete_tables_from_database(db_configuration=ChildDateBasedModel._db_configuration,
                                        table_name=_TABLE_NAME, schema_name=_SCHEMA_NAME)
        except Exception:
            pass

        DatabaseConnection.close_all_connections()
        DatabaseConnection.remove_all_connections()


if __name__ == '__main__':
    main()



"""Module contains transformer class for transforming empty strings to something else"""
from typing import Dict, Any

from .base import BaseTransformer
from ..utils.data import remove_empty_string_values


class EmptyStringTransformer(BaseTransformer):
    """Transformer class for transforming empty strings to some thing else"""
    replacement: Any = None

    @classmethod
    def run(cls, data: Dict[Any, Any]) -> Dict[Any, Any]:
        """Converts empty strings into some other value"""
        return {
            key: remove_empty_string_values(value=value, replacement=cls.replacement)
            for key, value in data.items()
        }



"""
Utility functions for selenium
https://towardsdatascience.com/using-python-and-selenium-to-automate-filling-forms-and-mouse-clicks-f87c74ed5c0f
"""
import os
import time

from datetime import timedelta, datetime

from selenium import webdriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from webdriver_manager.chrome import ChromeDriverManager

from .assets import get_csv_download_location

from pydantic import BaseModel


class WebDriverOptions(BaseModel):
    downloads_folder_location: str = get_csv_download_location()
    headless: bool = True


def get_web_driver(options: WebDriverOptions) -> webdriver.Chrome:
    """Returns a working instance of a chrome webdriver"""
    download_folder_location = options.downloads_folder_location

    chrome_options = webdriver.ChromeOptions()
    if options.headless:
        chrome_options.add_argument("--headless")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--disable-extensions")

    prefs = {
        "profile.default_content_settings.popups": 0,
        "download.default_directory": download_folder_location,
        "directory_upgrade": True
    }
    chrome_options.add_experimental_option('prefs', prefs)

    return webdriver.Chrome(
        executable_path=ChromeDriverManager().install(), options=chrome_options)


def visit_website(driver: webdriver.Chrome, website_url: str):
    """Visits a given website using selenium and the chromedriver"""
    return driver.get(url=website_url)


def get_html_element_by_xpath(driver: webdriver.Chrome, xpath: str, timeout: int = 2) -> WebElement:
    """Returns an instance of the element so as to manipulate it"""

    def is_html_element_visible(web_driver: webdriver.Chrome):
        """Returns the element if visible or False if not visible"""
        element = web_driver.find_element_by_xpath(xpath=xpath)
        return element if element else False

    return WebDriverWait(driver=driver, timeout=timeout).until(is_html_element_visible)


def wait_for_download_to_complete(expected_file_path: str, timeout: int = 10):
    """Waits for the download to complete"""
    start_time = datetime.now()
    timeout_delta = timedelta(seconds=timeout)

    while not os.path.exists(expected_file_path) and datetime.now() - start_time < timeout_delta:
        time.sleep(1)

    if not os.path.isfile(expected_file_path):
        raise IOError(f"{expected_file_path} failed to download in {timeout} seconds")


def wait_for_element_to_load(driver: webdriver, xpath: str, timeout: int):
    WebDriverWait(driver=driver, timeout=timeout).until(
        EC.presence_of_element_located((By.XPATH, xpath)))



"""Tests for the empty string transformer"""
from typing import Any
from unittest import TestCase, main

from judah.transformers.empty_string_transformer import EmptyStringTransformer


class TestEmptyStringTransformer(TestCase):
    """Tests for the EmptyStringTransformer base class"""

    def test_run(self):
        """
        should transform empty strings to some thing else
        """
        test_replacement = 'hello world'
        input_data = {
            'name': '',
            'address': 'Katakwi',
            'email': ''
        }

        expected_output_data = {
            'name': test_replacement,
            'address': 'Katakwi',
            'email': test_replacement
        }

        class ChildEmptyStringTransformer(EmptyStringTransformer):
            replacement: Any = test_replacement

        self.assertDictEqual(ChildEmptyStringTransformer.run(input_data), expected_output_data)


if __name__ == '__main__':
    main()



"""Base class for models"""
import logging
import sqlite3
from typing import Dict, Any, Type, List, Optional

import sqlalchemy
from sqlalchemy import inspect, and_, or_, Column, DateTime, schema, engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session
from sqlalchemy.exc import IntegrityError, ProgrammingError
from sqlalchemy.sql import func

from ..base import DestinationBaseModel
from .config import DatabaseConnectionConfig
from .connection import DatabaseConnection


class SlimDatabaseBaseModel(DestinationBaseModel):
    """
    The Slim Database Base Model without any columns predefined.
    """
    __table_args__: Dict = {}
    __table__: Any
    _db_configuration: DatabaseConnectionConfig
    _base_declarative_class: Type[declarative_base()]
    _datetime_fields: Optional[List[str]] = None

    @classmethod
    def get_attributes(cls) -> List[str]:
        """Returns the names of the columns so that they can be used, say in making an HTTP request"""
        return cls.__table__.columns.keys()

    @classmethod
    def get_last_record(cls):
        """Returns the last record that was put in the database"""
        if not isinstance(cls._datetime_fields, list):
            return None

        try:
            with DatabaseConnection.get_db_connection(
                    db_connection_config=cls._db_configuration) as db_connection:
                session = db_connection.db_session
                return session.query(cls).order_by(
                    *(getattr(cls, datetime_field).desc()
                      for datetime_field in cls._datetime_fields)).first()
        except sqlalchemy.exc.ProgrammingError:
            return None

    @classmethod
    def get_last_saved_timestamp(cls) -> Optional[List[Any]]:
        """Returns the timestamp of the last saved record"""
        last_record = cls.get_last_record()

        if isinstance(last_record, cls):
            timestamps = [getattr(last_record, datetime_field) for datetime_field in cls._datetime_fields]
            return timestamps[0] if len(timestamps) == 1 else timestamps

        return None

    def update(self, session: Session, **kwargs):
        """
        Updates the attributes passed in the kwargs and saves
        """
        for key, word in kwargs.items():
            self.__setattr__(key, word)
        self.save(session)

        return self

    def save(self, session: Session):
        """Commits changes to the database"""
        session.add(self)
        session.commit()

    def delete(self, session: Session, ):
        """Deletes the current record"""
        session.delete(self)
        session.commit()

    @classmethod
    def create_schema(cls, db_engine: engine.Engine):
        """Create the schema for this class if it exists"""
        schema_name = cls.__table_args__.get('schema', None)

        if schema_name is not None:
            try:
                db_engine.execute(schema.CreateSchema(schema_name))
            except (IntegrityError, ProgrammingError,):
                pass
            except (sqlite3.OperationalError, sqlalchemy.exc.OperationalError,) as exp:
                raise Exception(f'For sqlite databases, just use the "main" schema.'
                                f'The others are chaos!!!\n Original Exception: {exp}')

    @classmethod
    def initialize(cls):
        """Creates the tables in the database"""
        with DatabaseConnection.get_db_connection(
                db_connection_config=cls._db_configuration) as db_connection:
            try:
                db_engine = db_connection.connection_engine

                cls.create_schema(db_engine=db_engine)

                cls._base_declarative_class.metadata.create_all(bind=db_engine)
            except (IntegrityError, ProgrammingError) as exp:
                logging.info(f"{cls.__table__} already created. \n{exp}")

    @classmethod
    def upsert(cls, data: Dict[Any, Any]) -> Dict[Any, Any]:
        """
        Updates the given record row or creates it if it does not exist
        Returns data so that it can be used by the next pipe
        """
        with DatabaseConnection.get_db_connection(
                db_connection_config=cls._db_configuration) as db_connection:
            session = db_connection.db_session

            primary_key_columns = [column for column in inspect(cls).primary_key]
            is_data_with_primary_key = any([column.key in data for column in primary_key_columns])

            if is_data_with_primary_key:
                record = session.query(cls).filter(
                    or_(and_(*(column == data[column.key] for column in primary_key_columns)))
                ).first()

                if record:
                    record.update(session=session, **data)
                    return data

            record = cls(**data)
            record.save(session)
            return data


class DatabaseBaseModel(SlimDatabaseBaseModel):
    created_at = Column(DateTime(timezone=True), default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())



"""Utility functions for handling assets"""
import logging
import os
import shutil
from csv import DictReader
from datetime import datetime
from enum import Enum
from typing import Iterator, Dict, Any, Optional, List

import xlrd
from pydantic import BaseModel
from xml_stream import read_xml_file

ASSET_FOLDER_PATH = os.path.join(os.getcwd(), 'assets')


class FileType(Enum):
    CSV = 'csv'
    XLS = 'xls'
    XML = 'xml'


class FileOptions(BaseModel):
    """Options to pass the read_file function"""
    xml_records_tag: Optional[str] = None
    xls_sheet_name: Optional[str] = None
    xls_sheet_index: Optional[int] = 1
    xls_or_csv_headers: Optional[List[str]] = None


def get_asset_path(asset_name: Optional[str] = None):
    """Returns the path of a given asset"""
    if asset_name is None:
        return ASSET_FOLDER_PATH

    return os.path.join(ASSET_FOLDER_PATH, asset_name)


def get_timestamped_folder_name(
        prefix: str = '', suffix: str = '', date_format: str = '%m_%d_%Y_%H_%M_%S'):
    """Returns a folder name of format name"""
    now = datetime.utcnow()
    return f"{prefix}{now.strftime(date_format)}{suffix}"


def get_timestamped_folder(
        root_path: str, prefix: str = '', suffix: str = '', date_format: str = '%m_%d_%Y_%H_%M_%S'):
    """Returns a folder path with the given folder created if it did not exist"""
    timestamped_folder_name = get_timestamped_folder_name(prefix=prefix, suffix=suffix, date_format=date_format)
    timestamped_folder = os.path.join(root_path, timestamped_folder_name)

    if not os.path.exists(timestamped_folder):
        os.makedirs(timestamped_folder)

    return timestamped_folder


def get_csv_download_location(dataset_name: str = ''):
    """Returns the path to the csv folder asset"""
    csv_folder_path = get_asset_path(asset_name='csv')
    return get_timestamped_folder(root_path=csv_folder_path, prefix=dataset_name)


def get_xml_download_location(dataset_name: str = ''):
    """Returns the path to the xml folder asset"""
    csv_folder_path = get_asset_path(asset_name='xml')
    return get_timestamped_folder(root_path=csv_folder_path, prefix=dataset_name)


def read_csv_file(file_path: str, xls_or_csv_headers: Optional[List[str]] = None, **kwargs) -> Iterator[Dict[str, Any]]:
    """Reads the CSV file and returns the rows in the file as an iterator"""
    with open(file_path, 'r') as csv_file:
        csv_dict_reader = DictReader(csv_file, fieldnames=xls_or_csv_headers)
        yield from csv_dict_reader


def read_xls_file(file_path: str,
                  xls_or_csv_headers: Optional[List[str]] = None,
                  xls_sheet_name: Optional[str] = None,
                  xls_sheet_index: Optional[int] = 0, **kwargs) -> Iterator[Dict[str, Any]]:
    """Reads the XLS file and returns the rows in the file as an iterator"""
    with xlrd.open_workbook(file_path, on_demand=True) as xls_file:
        if isinstance(xls_sheet_name, str):
            sheet = xls_file.sheet_by_name(xls_sheet_name)
        elif isinstance(xls_sheet_index, int):
            sheet = xls_file.sheet_by_index(xls_sheet_index)
        else:
            raise ValueError('A sheet_name or sheet_index should be provided')

        for row in sheet.get_rows():
            row_values = [cell.value for cell in row]

            # should run once on the first row
            if xls_or_csv_headers is None:
                xls_or_csv_headers = row_values
                continue

            yield dict(zip(xls_or_csv_headers, row_values))


def read_file(file_path: str,
              file_type: FileType = FileType.CSV,
              options: FileOptions = FileOptions()) -> Iterator[Dict[str, Any]]:
    """Reads the Downloaded file and returns the rows in the file as an iterator"""
    with open(file_path, 'r') as file:
        if file_type == FileType.CSV:
            yield from read_csv_file(file_path=file_path, **dict(options))

        elif file_type == FileType.XLS:
            yield from read_xls_file(file_path=file_path, **dict(options))

        elif file_type == FileType.XML:
            yield from read_xml_file(
                file_path=file_path, to_dict=True, records_tag=options.xml_records_tag)

        else:
            raise ValueError('The given file_type cannot be handled by program')


def delete_parent_folder(file_path: str):
    """Deletes the parent folder when given a file path"""
    parent_folder = os.path.dirname(file_path)

    try:
        shutil.rmtree(parent_folder)
    except OSError as exp:
        logging.error(exp)



from datetime import date
from typing import Optional, Tuple, Type, Iterator, Dict, Any

from ..base import BaseController
from ...sources.base.quarter_based import QuarterBasedBaseSource
from ...utils.dates import convert_date_to_quarter_year_tuple


class QuarterBasedBaseController(BaseController):
    """Base controller class for quarter based data sources"""
    _source: QuarterBasedBaseSource
    _last_timestamp: Optional[date] = None

    start_quarter_and_year: Optional[Tuple[int, int]] = None
    end_quarter_and_year: Optional[Tuple[int, int]] = None
    source_class: Type[QuarterBasedBaseSource]

    @classmethod
    def _update_last_timestamp(cls):
        """Updates the last timestamp property of this class"""
        super()._update_last_timestamp()

        if isinstance(cls._last_timestamp, date):
            cls.start_quarter_and_year = convert_date_to_quarter_year_tuple(cls._last_timestamp)

    @classmethod
    def _query_source(cls, ) -> Iterator[Dict[Any, Any]]:
        """Queries the data source"""
        cls._update_source(start_quarter_and_year=cls.start_quarter_and_year,
                           end_quarter_and_year=cls.end_quarter_and_year)
        return cls._source.get()




"""Module with utility functions for logging"""
import logging
from logging.handlers import RotatingFileHandler


def setup_rotating_file_logger(logger: logging.Logger, file_path: str, max_bytes: int = 2000000,
                               log_format: str = '%(asctime)-15s \n%(name)s - %(levelname)s\n\n %(message)s\n\n'):
    """
    Makes the logger to use a file that rotates when a given size is reached.
    It has a single back up and logs errors and above
    """
    logger_handler = RotatingFileHandler(file_path, maxBytes=max_bytes, backupCount=1)
    logger_formatter = logging.Formatter(log_format)
    logger_handler.setFormatter(logger_formatter)
    logger_handler.setLevel(logging.ERROR)

    logger.addHandler(logger_handler)
    return logger



"""Module containing tests for the dates utility functions"""
from datetime import datetime, date
from typing import Any
from unittest import TestCase, main
from unittest.mock import patch

import pytz

from judah.utils.dates import (
    convert_date_to_string,
    get_default_start_year,
    get_default_start_month,
    get_quarter_from_month,
    get_default_start_quarter,
    get_current_year,
    get_current_quarter,
    get_default_historical_start_date,
    get_default_historical_start_quarter_and_year,
    update_date,
    update_quarter_year_tuple,
    convert_date_to_quarter_year_tuple,
    change_datetime_format
)


class TestDatesUtilities(TestCase):
    """Tests for the dates utility functions"""

    def setUp(self) -> None:
        """Initialize some variables"""
        self.month = 11
        self.day = 11
        self.year = 2020
        self.hour = 22
        self.minute = 10
        self.second = 33
        self.timezone_name = "Africa/Kampala"
        self.utc_offset_hours = 3
        self.timezone = pytz.timezone(self.timezone_name)
        self.datetime_format = "%Y/%m/%dT%H:%M:%S%z"
        self.month_quarter_map = {1: 1, 2: 1, 3: 1, 4: 2, 5: 2, 6: 2, 7: 3, 8: 3, 9: 3, 10: 4, 11: 4, 12: 4}

    def test_convert_date_to_string(self):
        """Converts a date to the expected string format defaulting to UTC timezone and no timezone separator"""
        test_datetime = datetime(
            year=self.year, month=self.month,
            day=self.day,
            tzinfo=pytz.utc)
        date_string = f'{self.year}/{self.month}/{self.day}'
        expected_datetime_string = f'{date_string}T00:00:00+0000'

        self.assertEqual(convert_date_to_string(test_datetime, self.datetime_format), expected_datetime_string)

    def test_convert_date_to_string_with_timezone(self):
        """Converts a date to the expected string format with given timezone and no timezone separator"""
        test_datetime = datetime(
            year=self.year, month=self.month,
            day=self.day,
            tzinfo=pytz.utc).astimezone(self.timezone)
        date_string = f'{self.year}/{self.month}/{self.day}'
        utc_offset_hours_string = f'{self.utc_offset_hours}'.zfill(2)
        expected_datetime_string = f'{date_string}T00:00:00+{utc_offset_hours_string}00'

        returned_datetime_string = convert_date_to_string(
            test_datetime, self.datetime_format, timezone_name=self.timezone_name)

        self.assertEqual(returned_datetime_string, expected_datetime_string)

    def test_convert_date_to_string_with_timezone_separator(self):
        """Converts a date to the expected string format defaulting to UTC timezone and given timezone separator"""
        test_datetime = datetime(
            year=self.year, month=self.month,
            day=self.day,
            tzinfo=pytz.utc)
        date_string = f'{self.year}/{self.month}/{self.day}'
        timezone_separator = ':'
        expected_datetime_string = f'{date_string}T00:00:00+00{timezone_separator}00'

        returned_datetime_string = convert_date_to_string(
            test_datetime, self.datetime_format, timezone_separator=timezone_separator)

        self.assertEqual(returned_datetime_string, expected_datetime_string)

    @patch('os.getenv')
    def test_get_default_start_year(self, mock_os_getenv):
        """
        Should return the HISTORICAL_STARTING_YEAR from the environment
        """
        test_start_year = 1999
        mock_os_getenv.return_value = f'{test_start_year}'

        self.assertEqual(get_default_start_year(), test_start_year)
        mock_os_getenv.assert_called_once_with('HISTORICAL_STARTING_YEAR', None)

    @patch('os.getenv')
    def test_get_default_start_month(self, mock_os_getenv):
        """
        Should return the HISTORICAL_STARTING_MONTH from the environment
        """
        test_start_month = 5
        mock_os_getenv.return_value = f'{test_start_month}'

        self.assertEqual(get_default_start_month(), test_start_month)
        mock_os_getenv.assert_called_once_with('HISTORICAL_STARTING_MONTH', None)

    @patch('judah.utils.dates.get_default_start_month')
    def test_get_default_start_quarter(self, mock_get_default_start_month):
        """
        Should return the quarter in which the given default start month belongs to
        """
        for month, quarter in self.month_quarter_map.items():
            mock_get_default_start_month.return_value = month
            self.assertEqual(get_default_start_quarter(), quarter)

    def test_get_quarter_from_month(self):
        """
        Should return the quarter in which the given start month belongs to
        """
        for month, quarter in self.month_quarter_map.items():
            self.assertEqual(get_quarter_from_month(month), quarter)

    def test_get_current_year(self):
        """
        Should return the current year
        """
        current_year = datetime.now().year
        self.assertEqual(get_current_year(), current_year)

    def test_get_current_quarter(self):
        """
        Should return the quarter in which the current month is in
        """
        current_month = datetime.now().month
        current_quarter = self.month_quarter_map[current_month]
        self.assertEqual(get_current_quarter(), current_quarter)

    @patch('os.getenv')
    def test_get_default_historical_start_date(self, mock_os_getenv):
        """
        Should return the date object got from
        HISTORICAL_STARTING_YEAR, HISTORICAL_STARTING_MONTH and HISTORICAL_STARTING_DAY
        """
        test_year = 1998
        test_month = 4
        test_day = 7

        def handle_os_getenv(string: str, default: Any):
            return_map = {
                'HISTORICAL_STARTING_YEAR': f'{test_year}',
                'HISTORICAL_STARTING_MONTH': f'{test_month}',
                'HISTORICAL_STARTING_DAY': f'{test_day}'
            }
            return return_map[string]

        mock_os_getenv.side_effect = handle_os_getenv
        expected_date = date(year=test_year, month=test_month, day=test_day)

        self.assertEqual(get_default_historical_start_date(), expected_date)

    @patch('os.getenv')
    def test_get_default_historical_start_quarter_and_year(self, mock_os_getenv):
        """
        Should return the quarter, year tuple got from
        HISTORICAL_STARTING_YEAR, HISTORICAL_STARTING_MONTH
        """
        test_year = 1997
        test_month = 7

        def handle_os_getenv(string: str, default: Any):
            return_map = {
                'HISTORICAL_STARTING_YEAR': f'{test_year}',
                'HISTORICAL_STARTING_MONTH': f'{test_month}',
            }
            return return_map[string]

        mock_os_getenv.side_effect = handle_os_getenv
        expected_quarter_and_month = (self.month_quarter_map[test_month], test_year,)

        self.assertEqual(get_default_historical_start_quarter_and_year(), expected_quarter_and_month)

    def test_update_date(self):
        """
        Should increment or decrement date accordingly
        """
        test_date = date(year=2020, month=3, day=8)
        days_to_increment_by = 31
        days_to_decrement_by = 8

        last_day_of_february = date(year=2020, month=2, day=29)
        test_date_after_31_days = date(year=2020, month=4, day=8)
        test_date_after_23_days = date(year=2020, month=3, day=31)

        incremented_date = update_date(
            test_date, days_to_decrement_by=0, days_to_increment_by=days_to_increment_by)
        decremented_date = update_date(
            test_date, days_to_decrement_by=days_to_decrement_by, days_to_increment_by=0)
        incremented_and_decremented_date = update_date(
            test_date, days_to_decrement_by=days_to_decrement_by, days_to_increment_by=days_to_increment_by)

        self.assertEqual(decremented_date, last_day_of_february)
        self.assertEqual(incremented_date, test_date_after_31_days)
        self.assertEqual(incremented_and_decremented_date, test_date_after_23_days)

    def test_update_quarter_year_tuple(self):
        """
        Should increment or decrement a (quarter, year,) tuple accordingly
        """
        test_quarter_year_tuple = (3, 2020,)
        quarters_to_increment_by = 4
        quarters_to_decrement_by = 3

        last_quarter_year_tuple_of_2019 = (4, 2019,)
        test_quarter_year_tuple_after_4_quarters = (3, 2021,)
        test_quarter_year_tuple_after_1_quarter = (4, 2020,)

        incremented_quarter_year_tuple = update_quarter_year_tuple(
            test_quarter_year_tuple, quarters_to_increment_by=quarters_to_increment_by, quarters_to_decrement_by=0)
        decremented_quarter_year_tuple = update_quarter_year_tuple(
            test_quarter_year_tuple, quarters_to_increment_by=0, quarters_to_decrement_by=quarters_to_decrement_by)
        incremented_and_decremented_quarter_year_tuple = update_quarter_year_tuple(
            test_quarter_year_tuple, quarters_to_increment_by=quarters_to_increment_by,
            quarters_to_decrement_by=quarters_to_decrement_by)

        self.assertEqual(incremented_quarter_year_tuple, test_quarter_year_tuple_after_4_quarters)
        self.assertEqual(decremented_quarter_year_tuple, last_quarter_year_tuple_of_2019)
        self.assertEqual(incremented_and_decremented_quarter_year_tuple, test_quarter_year_tuple_after_1_quarter)

    def test_convert_date_to_quarter_year_tuple(self):
        """
        Should convert a given date to a quarter_year tuple
        """
        date_quarter_map = {
            date(year=2020, month=3, day=20): 1,
            date(year=2010, month=5, day=2): 2,
            date(year=1990, month=7, day=30): 3,
            date(year=2020, month=10, day=10): 4,
        }

        for test_date, quarter in date_quarter_map.items():
            self.assertEqual(convert_date_to_quarter_year_tuple(test_date), (quarter, test_date.year))

    def test_change_datetime_format(self):
        """
        Should convert a date string's format to another format
        """
        original_format = '%Y-%m-%d %H:%M:%S%z'
        new_format = '%Y/%d/%m/%dT%H-%M-%S%Z'
        original_date_string = '2002-04-03 12:38:42+0000'
        expected_date_string = '2002/03/04/03T12-38-42UTC'

        returned_date_string = change_datetime_format(
            datetime_value=original_date_string, old=original_format, new=new_format)

        self.assertEqual(returned_date_string, expected_date_string)


if __name__ == '__main__':
    main()



"""Module with tests for the date based database source source"""
import os
from collections import Iterator
from datetime import date
from typing import Dict, Type, Optional, List, Any
from unittest import TestCase, main

from dotenv import load_dotenv
from sqlalchemy.ext.declarative import declarative_base
import sqlalchemy as orm

from judah.destinations.database.config import DatabaseConnectionConfig
from judah.destinations.database.connection import DatabaseConnection
from judah.destinations.database.model import DatabaseBaseModel
from judah.sources.database.config import DatabaseSourceDataModel
from judah.sources.database.date_based import DateBasedDatabaseSource
from test.utils import delete_tables_from_database, create_tables_in_database

_ROOT_PATH = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

load_dotenv(os.path.join(_ROOT_PATH, '.env'))

_TEST_DB_URI = os.environ.get('TEST_POSTGRES_DB_URI')

_TEST_DB_BASE = declarative_base()
_TABLE_NAME = 'dummy_source'
_SCHEMA_NAME = 'test_schema_for_source'


class TestModel(DatabaseBaseModel, _TEST_DB_BASE):
    """Test database model"""
    __tablename__ = _TABLE_NAME
    __table_args__: Dict = {'schema': _SCHEMA_NAME}
    _db_configuration: DatabaseConnectionConfig = DatabaseConnectionConfig(db_uri=_TEST_DB_URI)
    _base_declarative_class: Type[declarative_base()] = _TEST_DB_BASE
    _datetime_fields: Optional[List[str]] = ["Date"]

    Date = orm.Column(orm.Date, primary_key=True)
    number = orm.Column(orm.Integer, primary_key=True)
    Capacity = orm.Column(orm.Integer)
    Price = orm.Column(orm.Integer)


class TestDatabaseSourceDataModel(DatabaseSourceDataModel):
    Date: date
    number: int
    Capacity: int
    Price: int


class TestDateBasedDatabaseSource(TestCase):
    """Tests for the DateBasedDatabaseSource"""

    def setUp(self) -> None:
        """Initialize some variables"""
        self.data = [
            {"Date": date(year=2020, month=3, day=9), "number": 1, "Capacity": 16616, "Price": 67},
            {"Date": date(year=2020, month=3, day=12), "number": 2, "Capacity": 16516, "Price": 567},
            {"Date": date(year=2020, month=3, day=10), "number": 3, "Capacity": 16616, "Price": 637},
            {"Date": date(year=2020, month=3, day=9), "number": 4, "Capacity": 16620, "Price": 617},
        ]
        try:
            delete_tables_from_database(db_configuration=TestModel._db_configuration,
                                        table_name=_TABLE_NAME, schema_name=_SCHEMA_NAME)
        except Exception:
            pass

    def load_database(self):
        """Loads the database with the self.data"""
        create_tables_in_database(db_configuration=TestModel._db_configuration,
                                  model_base=_TEST_DB_BASE, schema_name=_SCHEMA_NAME)

        with DatabaseConnection.get_db_connection(
                db_connection_config=TestModel._db_configuration) as db_connection:
            for datum in self.data:
                record = TestModel(**datum)
                db_connection.db_session.add(record)

            db_connection.db_session.commit()

    def test_get_with_params(self):
        """Makes an SQL query with the params from the _get_params method and returns a generator (iterator)"""
        self.load_database()

        dummy_params = {'number': 3}

        class DateBasedDatabaseSourceWithParams(DateBasedDatabaseSource):
            sql_query: str = f"""
            SELECT * FROM {_SCHEMA_NAME}.{_TABLE_NAME}
            WHERE "number" = :number
            """
            name: str = 'sample'
            db_configuration: DatabaseConnectionConfig = TestModel._db_configuration
            data_model_class: Type[DatabaseSourceDataModel] = TestDatabaseSourceDataModel

            def _get_params(self, start_date: Optional[date] = None, end_date: Optional = None) -> Dict[Any, Any]:
                return dummy_params

        source = DateBasedDatabaseSourceWithParams()
        response = source.get()

        self.assertIsInstance(response, Iterator)

        expected_data = [self.data[2]]
        received_data = list(response)

        self.assertListEqual(expected_data, received_data)

    def test_get_with_default_params(self):
        """Makes an SQL query with the default params of {} and returns an appropriate generator (iterator)"""
        self.load_database()

        class DateBasedDatabaseSourceWithParams(DateBasedDatabaseSource):
            sql_query: str = f"""
            SELECT * FROM {_SCHEMA_NAME}.{_TABLE_NAME}
            WHERE "number" IN (2, 4)
            """
            name: str = 'sample'
            db_configuration: DatabaseConnectionConfig = TestModel._db_configuration
            data_model_class: Type[DatabaseSourceDataModel] = TestDatabaseSourceDataModel

        source = DateBasedDatabaseSourceWithParams()
        response = source.get()

        self.assertIsInstance(response, Iterator)

        expected_data = [self.data[1], self.data[3]]
        expected_data.sort(key=lambda x: x['number'])

        received_data = list(response)
        received_data.sort(key=lambda x: x['number'])

        self.assertListEqual(expected_data, received_data)

    def tearDown(self) -> None:
        """Clean up after the tests"""
        try:
            delete_tables_from_database(db_configuration=TestModel._db_configuration,
                                        table_name=_TABLE_NAME, schema_name=_SCHEMA_NAME)
        except Exception:
            pass

        DatabaseConnection.close_all_connections()
        DatabaseConnection.remove_all_connections()


if __name__ == '__main__':
    main()



"""Module containing the controller class for getting data from the quarter-based export site to the database"""

from typing import Type, Optional, Tuple

from ..base.quarter_based import QuarterBasedBaseController
from ...destinations.database.model import DatabaseBaseModel
from ...sources.export_site.quarter_based import QuarterBasedExportSiteSource
from ...utils.dates import get_default_historical_start_quarter_and_year


class QuarterBasedExportSiteToDatabaseController(QuarterBasedBaseController):
    """
    Class for the controller that downloads data from the quarter-based export site
    and saves it in a database
    """
    _source: QuarterBasedExportSiteSource

    destination_model_class: Type[DatabaseBaseModel]
    source_class: Type[QuarterBasedExportSiteSource]
    start_quarter_and_year: Optional[Tuple[int, int]] = get_default_historical_start_quarter_and_year()


"""Module containing configurations for the date-based REST API data source"""

from datetime import date
from typing import Iterator, Dict, Any, Optional

import requests

from ..base.date_based import DateBasedBaseSource
from ...utils.dates import convert_date_to_string


class RequestError(BaseException):
    """Exception that occurs when a request fails"""

    def __init__(self, status_code: int, message: Optional[str] = ''):
        self.status_code = status_code
        self.message = message
        super().__init__(f'{status_code}: {self.message}')


class DateBasedRestAPISource(DateBasedBaseSource):
    """Base class for the REST API source that depends on dates"""
    response_data_key: Optional[str] = None
    default_headers: Optional[Dict[Any, Any]] = {}
    start_date_param_name: Optional[str] = None
    end_date_param_name: Optional[str] = None
    date_format: str = '%Y-%m-%d'
    timezone: Optional[str] = None

    def _authenticate(self):
        """A method to generate headers for authorization and update the default_headers"""
        pass

    def _get_headers(self) -> Dict[Any, Any]:
        """An overridable method to return headers use in the request"""
        return self.default_headers.copy()

    def _get_params(self, **kwargs) -> Dict[Any, Any]:
        """An overridable method to return query params dict to use in the request"""
        params = {**kwargs}

        if isinstance(self.start_date_param_name, str):
            start_date: Optional[date] = params.get(self.start_date_param_name, None)

            if start_date is not None:
                params[self.start_date_param_name] = convert_date_to_string(
                    date_value=start_date, date_format=self.date_format, timezone_name=self.timezone)

        if isinstance(self.end_date_param_name, str):
            end_date: Optional[date] = params.get(self.end_date_param_name, None)

            if end_date is not None:
                params[self.end_date_param_name] = convert_date_to_string(
                    date_value=end_date, date_format=self.date_format, timezone_name=self.timezone)

        return params

    def _get_data_url(self):
        """Gets the url for given data from the REST API"""
        if self.base_uri is None or self.name is None:
            raise Exception("base_uri and name attributes should not be None")

        return f"{self.base_uri}/{self.name}"

    def _query_url(self, url: str, start_date: date, end_date: date) -> requests.Response:
        """Make a requests GET request for a given set of dates"""
        headers = self._get_headers()
        params = self._get_params(**{self.start_date_param_name: start_date, self.end_date_param_name: end_date})
        return requests.get(url=url, headers=headers, params=params)

    def _extract_list_from_response(self, response: requests.Response):
        """Extracts the list data from the requests.Response object"""
        json_response = response.json()

        if self.response_data_key is not None:
            json_response = json_response.get(self.response_data_key)

        if isinstance(json_response, dict):
            yield json_response
        else:
            yield from json_response

    def _query_data_source(self, start_date: date, end_date: date) -> Iterator[Dict[str, Any]]:
        """Queries a given start and end date and returns an iterator with data records"""
        url = self._get_data_url()
        response = self._query_url(url=url, start_date=start_date, end_date=end_date)

        if response.status_code in (401, 403, 400,):
            self._authenticate()
            response = self._query_url(url=url, start_date=start_date, end_date=end_date)

        if response.ok:
            yield from self._extract_list_from_response(response=response)
        else:
            raise RequestError(message=response.text, status_code=response.status_code)



"""Module containing tests for the data utility functions"""

from unittest import TestCase, main
from judah.utils.data import remove_empty_string_values, replace_keys_with_aliases, replace_substrings


class TestDataUtilities(TestCase):
    """Tests for the data utility functions"""

    def test_remove_empty_string_values(self):
        """Removes empty string value, replacing it with a given replacement"""
        replacement = 'hello'
        self.assertEqual(remove_empty_string_values('', replacement=replacement), replacement)
        self.assertEqual(remove_empty_string_values(''), None)
        self.assertEqual(remove_empty_string_values('None empty', replacement=replacement), 'None empty')

    def test_replace_keys_with_aliases(self):
        """Replaces keys in a dict with aliases from a key-alias map"""
        alias_map = {
            'nom': 'name',
            'adresse': 'address'
        }
        data = [
            {'nom': 'John Doe', 'adresse': 'Hoima'},
            {'nom': 'Jean Doe', 'adresse': 'Paris'},
            {'nom': 'Nazir Doe', 'adresse': 'Cairo'},
        ]

        expect_output = [
            {'name': 'John Doe', 'address': 'Hoima'},
            {'name': 'Jean Doe', 'address': 'Paris'},
            {'name': 'Nazir Doe', 'address': 'Cairo'},
        ]

        for datum, expected_datum in zip(data, expect_output):
            processed_datum = replace_keys_with_aliases(alias_map=alias_map, data=datum)
            self.assertDictEqual(processed_datum, expected_datum)

    def test_replace_substrings(self):
        """Replaces characters in a given string with other characters"""
        substring_replacement_map = {
            'foo': 'bar',
            'tea': 'java'
        }
        phrase = '"foo" is just some word people use. It is like "tea"'
        expected_phrase = '"bar" is just some word people use. It is like "java"'

        processed_phrase = replace_substrings(substring_replacement_map=substring_replacement_map, string=phrase)

        self.assertEqual(processed_phrase, expected_phrase)


if __name__ == '__main__':
    main()



"""Module containing the base transformer for data"""
from typing import Dict, Any, Union, List


class BaseTransformer:
    """
    The base class for all data transformers that receive a dictionary
    and return a transformed dictionary
    """

    @classmethod
    def run(cls, data: Dict[Any, Any]) -> Union[Dict[Any, Any], List[Any]]:
        """This transforms the data"""
        return data



"""Module for transforming the data field names by removing unwanted characters"""
from typing import Dict, Any

from .base import BaseTransformer
from ..utils.data import replace_substrings


class FieldNameCleanerTransformer(BaseTransformer):
    """Removes unwanted characters, replacing them with other characters in the field names"""
    substring_replacement_map: Dict[str, str] = {}

    @classmethod
    def run(cls, data: Dict[Any, Any]) -> Dict[Any, Any]:
        """Replaces unwanted characters with wanted characters"""
        return {
            replace_substrings(substring_replacement_map=cls.substring_replacement_map, string=str(key)): value
            for key, value in data.items()
        }



"""Module containing tests for the quarter based Exports Site data source"""
import os
from collections import Iterator
from datetime import date
from typing import Optional
from unittest import TestCase, main
from unittest.mock import patch, Mock, call

from selenium import webdriver

from judah.sources.export_site.quarter_based import QuarterBasedExportSiteSource
from judah.utils.dates import convert_date_to_quarter_year_tuple, update_quarter_year_tuple

_PARENT_FOLDER = os.path.dirname(__file__)
_MOCK_ASSET_FOLDER_PATH = os.path.join(os.path.dirname(os.path.dirname(_PARENT_FOLDER)), 'assets')


class ChildQuarterBasedExportSiteSource(QuarterBasedExportSiteSource):
    """Child export site source that just picks a file from the file system"""
    base_uri: str = 'http://example.com'
    name: str = 'test_export_site_source'

    def _download_file(self, start_date: date, end_date: date) -> Optional[str]:
        """Downloads the CSV from the export site and returns the path to it"""
        return None


class TestQuarterBasedExportSiteSource(TestCase):
    """Tests for the QuarterBasedExportSiteSource"""

    def setUp(self) -> None:
        """Initialize some variables"""
        self.mock_csv_file_path = os.path.join(_MOCK_ASSET_FOLDER_PATH, 'mock.csv')
        self.expected_data = [
            {"Date": "09/03/2020", "number": "1", "period_from": "00:00", "period_until": "00:15", "Capacity": "16616"},
            {"Date": "09/03/2020", "number": "2", "period_from": "00:15", "period_until": "00:30", "Capacity": "16616"},
            {"Date": "09/03/2020", "number": "3", "period_from": "00:30", "period_until": "00:45", "Capacity": "16616"},
            {"Date": "09/03/2020", "number": "4", "period_from": "00:45", "period_until": "01:00", "Capacity": "16620"},
        ]

    @patch('judah.sources.export_site.quarter_based.visit_website')
    def test_initialize_chrome(self, mock_visit_website):
        """
        Should initialize Chrome in case it is not yet initialized and visits the base url
        """
        quarter_based_export_site_source = ChildQuarterBasedExportSiteSource()
        self.assertIsNone(quarter_based_export_site_source.chrome)

        quarter_based_export_site_source._initialize_chrome()

        self.assertIsInstance(quarter_based_export_site_source.chrome, webdriver.Chrome)
        mock_visit_website.assert_called_once_with(
            driver=quarter_based_export_site_source.chrome, website_url=quarter_based_export_site_source.base_uri)

        quarter_based_export_site_source.chrome.close()
        quarter_based_export_site_source.chrome.quit()

    @patch.object(ChildQuarterBasedExportSiteSource, '_initialize_chrome')
    def test_del(self, mock_initialize_chrome):
        """
        Should quit chrome on deletion
        """
        quarter_based_export_site_source = ChildQuarterBasedExportSiteSource()
        quarter_based_export_site_source.chrome = Mock(spec=webdriver.Chrome)
        quarter_based_export_site_source.__del__()
        quarter_based_export_site_source.chrome.quit.assert_called_once()

    @patch('judah.sources.export_site.quarter_based.delete_parent_folder')
    @patch.object(ChildQuarterBasedExportSiteSource, '_initialize_chrome')
    @patch.object(ChildQuarterBasedExportSiteSource, '_download_file')
    def test_query_data_source(self, mock_download_file, mock_initialize_chrome, mock_delete_parent_folder):
        """
        Should query a given start_quarter_and_year and end_quarter_and_year
        and return an iterator with data records and then deletes folder
        """
        # initializations
        mock_download_file.return_value = self.mock_csv_file_path
        quarter_based_export_site_source = ChildQuarterBasedExportSiteSource()
        start_quarter_and_year = (1, 2020,)
        end_quarter_and_year = (3, 2020,)

        # method call
        response = quarter_based_export_site_source._query_data_source(
            start_quarter_and_year=start_quarter_and_year, end_quarter_and_year=end_quarter_and_year)
        self.assertIsInstance(response, Iterator)

        data = [record for record in response]

        # assertions
        mock_initialize_chrome.assert_called_once()
        mock_download_file.assert_called_once_with(
            start_quarter_and_year=start_quarter_and_year, end_quarter_and_year=end_quarter_and_year)
        mock_delete_parent_folder.assert_called_once_with(self.mock_csv_file_path)
        self.assertListEqual(data, self.expected_data)

    @patch('judah.sources.export_site.quarter_based.delete_parent_folder')
    @patch.object(ChildQuarterBasedExportSiteSource, '_initialize_chrome')
    @patch.object(ChildQuarterBasedExportSiteSource, '_download_file')
    def test_query_data_source_no_file_downloaded(self, mock_download_file, mock_initialize_chrome,
                                                  mock_delete_parent_folder):
        """
        Should query a given start_quarter_and_year and end_quarter_and_year
         and return an empty iterator if there is no file downloaded
        """
        # initializations
        mock_download_file.return_value = None
        quarter_based_export_site_source = ChildQuarterBasedExportSiteSource()
        start_quarter_and_year = (1, 2020,)
        end_quarter_and_year = (3, 2020,)

        # method call
        response = quarter_based_export_site_source._query_data_source(
            start_quarter_and_year=start_quarter_and_year, end_quarter_and_year=end_quarter_and_year)
        self.assertIsInstance(response, Iterator)

        data = [record for record in response]

        # assertions
        mock_initialize_chrome.assert_called_once()
        mock_download_file.assert_called_once_with(
            start_quarter_and_year=start_quarter_and_year, end_quarter_and_year=end_quarter_and_year)
        mock_delete_parent_folder.assert_not_called()
        self.assertListEqual(data, [])

    @patch.object(ChildQuarterBasedExportSiteSource, '_query_data_source')
    @patch.object(ChildQuarterBasedExportSiteSource, '_get_next_end_quarter_and_year')
    @patch.object(ChildQuarterBasedExportSiteSource, '_get_next_start_quarter_and_year')
    def test_get(self, mock_get_next_start_quarter_and_year,
                 mock_get_next_end_quarter_and_year, mock_query_data_source):
        """
        Should return data from a given date to a given date as an iterator
        """
        start_quarter_and_year = (1, 2020,)
        end_quarter_and_year = (3, 2020,)
        mock_get_next_start_quarter_and_year.return_value = start_quarter_and_year
        mock_get_next_end_quarter_and_year.return_value = end_quarter_and_year
        mock_query_data_source.return_value = (record for record in self.expected_data)

        quarter_based_export_site_source = ChildQuarterBasedExportSiteSource()
        response = quarter_based_export_site_source.get()

        self.assertIsInstance(response, Iterator)

        data = [record for record in response]

        self.assertListEqual(data, self.expected_data)
        mock_get_next_start_quarter_and_year.assert_has_calls(
            [call(), call(quarters_to_increment_by=quarter_based_export_site_source.default_batch_size_in_quarters)])
        mock_get_next_end_quarter_and_year.assert_called_once()

    @patch.object(ChildQuarterBasedExportSiteSource, '_query_data_source')
    @patch.object(ChildQuarterBasedExportSiteSource, '_get_next_end_quarter_and_year')
    @patch.object(ChildQuarterBasedExportSiteSource, '_get_next_start_quarter_and_year')
    def test_get_end_quarter_earlier_than_start_quarter(self, mock_get_next_start_quarter_and_year,
                                                        mock_get_next_end_quarter_and_year, mock_query_data_source):
        """
        Should return an iterator of an empty list
        if the end_quarter_and_year is earlier than start_quarter_and_year
        """
        start_quarter_and_year = (1, 2020,)
        end_quarter_and_year = (3, 2019,)
        mock_get_next_start_quarter_and_year.return_value = start_quarter_and_year
        mock_get_next_end_quarter_and_year.return_value = end_quarter_and_year

        quarter_based_export_site_source = ChildQuarterBasedExportSiteSource()
        response = quarter_based_export_site_source.get()

        self.assertIsInstance(response, Iterator)

        data = [record for record in response]

        self.assertListEqual(data, [])
        mock_get_next_start_quarter_and_year.assert_called_once()
        mock_get_next_end_quarter_and_year.assert_called_once()
        mock_query_data_source.assert_not_called()

    def test_get_next_start_quarter_and_year_with_no_initial_quarter_year_tuples(self):
        """
        Should get the next start_quarter_and_year tuple as current quarter, year
         minus the default_batch_size_in_quarters plus net quarter increment
        when only default_batch_size_in_quarters is given
        """
        today = date.today()
        current_quarter_and_year_tuple = convert_date_to_quarter_year_tuple(today)
        quarters_to_increment_by = 4
        quarters_to_decrement_by = 8

        quarter_based_export_site_source = ChildQuarterBasedExportSiteSource()
        net_quarters_to_increment_by = (
                    quarters_to_increment_by
                    - quarters_to_decrement_by
                    - quarter_based_export_site_source.default_batch_size_in_quarters)

        next_start_quarter_and_year = quarter_based_export_site_source._get_next_start_quarter_and_year(
            quarters_to_increment_by=quarters_to_increment_by, quarters_to_decrement_by=quarters_to_decrement_by)

        expected_next_start_quarter_and_year = update_quarter_year_tuple(
            current_quarter_and_year_tuple, quarters_to_increment_by=net_quarters_to_increment_by,
            quarters_to_decrement_by=0)
        self.assertTupleEqual(next_start_quarter_and_year, expected_next_start_quarter_and_year)

    def test_get_next_start_quarter_and_year_given_end_quarter_and_year(self):
        """
        Should get the next start_quarter_and_year tuple as end_quarter_and_year tuple
        minus the default_batch_size_in_quarters
        plus net quarter increment  when default_batch_size_in_quarters and end_quarter_and_year_tuple are given
        """
        end_quarter_and_year = (2, 2019,)
        quarters_to_increment_by = 4
        quarters_to_decrement_by = 8

        quarter_based_export_site_source = ChildQuarterBasedExportSiteSource()
        quarter_based_export_site_source.end_quarter_and_year = end_quarter_and_year
        net_quarters_to_increment_by = (
                    quarters_to_increment_by
                    - quarters_to_decrement_by
                    - quarter_based_export_site_source.default_batch_size_in_quarters)

        next_start_quarter_and_year = quarter_based_export_site_source._get_next_start_quarter_and_year(
            quarters_to_increment_by=quarters_to_increment_by, quarters_to_decrement_by=quarters_to_decrement_by)

        expected_next_start_quarter_and_year = update_quarter_year_tuple(
            end_quarter_and_year, quarters_to_increment_by=net_quarters_to_increment_by,
            quarters_to_decrement_by=0)
        self.assertTupleEqual(next_start_quarter_and_year, expected_next_start_quarter_and_year)

    def test_get_next_start_quarter_and_year_with_initial_start_quarter_and_year(self):
        """
        Should get the next start_quarter_and_year tuple as start_quarter_and_year
         plus net quarter increment when start_quarter_and_year is given,
        regardless of end_quarter_and_year and default_batch_size_in_quarters
        """
        end_quarter_and_year = (1, 2019)
        start_quarter_and_year = (3, 2020,)
        quarters_to_increment_by = 4
        quarters_to_decrement_by = 15
        net_quarters_to_increment_by = quarters_to_increment_by - quarters_to_decrement_by

        quarter_based_export_site_source = ChildQuarterBasedExportSiteSource()
        quarter_based_export_site_source.end_quarter_and_year = end_quarter_and_year
        quarter_based_export_site_source.start_quarter_and_year = start_quarter_and_year

        next_start_quarter_and_year = quarter_based_export_site_source._get_next_start_quarter_and_year(
            quarters_to_increment_by=quarters_to_increment_by, quarters_to_decrement_by=quarters_to_decrement_by)

        expected_next_start_quarter_and_year = update_quarter_year_tuple(
            start_quarter_and_year, quarters_to_increment_by=net_quarters_to_increment_by,
            quarters_to_decrement_by=0)
        self.assertTupleEqual(next_start_quarter_and_year, expected_next_start_quarter_and_year)

    def test_get_next_end_quarter_and_year_with_no_initial_quarter_year_tuples(self):
        """
        Should get the next end_quarter_and_year as current quarter_and_year tuple plus net quarter increment
        when only default_batch_size_in_quarters is given
        Assumption: current quarter_and_year tuple is end_quarter_and_year if no increment or decrement
        """
        today = date.today()
        current_quarter_and_year_tuple = convert_date_to_quarter_year_tuple(today)
        quarters_to_increment_by = 4
        quarters_to_decrement_by = 9
        net_quarters_to_increment_by = quarters_to_increment_by - quarters_to_decrement_by

        quarter_based_export_site_source = ChildQuarterBasedExportSiteSource()
        next_end_quarter_and_year = quarter_based_export_site_source._get_next_end_quarter_and_year(
            quarters_to_increment_by=quarters_to_increment_by, quarters_to_decrement_by=quarters_to_decrement_by)

        expected_next_end_quarter_and_year = update_quarter_year_tuple(
            current_quarter_and_year_tuple, quarters_to_increment_by=net_quarters_to_increment_by,
            quarters_to_decrement_by=0)
        self.assertTupleEqual(next_end_quarter_and_year, expected_next_end_quarter_and_year)

    def test_get_next_end_quarter_and_year_given_start_quarter_and_year(self):
        """
        Should get the next end quarter_and_year as start_quarter_and_year plus the net quarter increment
        plus default_batch_size_in_quarters
        when default_batch_size_in_quarters and start_quarter_and_year are given and no initial end_quarter_and_year is given
        """
        start_quarter_and_year = (2, 2020)
        quarters_to_increment_by = 4
        quarters_to_decrement_by = 9

        quarter_based_export_site_source = ChildQuarterBasedExportSiteSource()
        quarter_based_export_site_source.start_quarter_and_year = start_quarter_and_year
        net_quarters_to_increment_by = (
                quarters_to_increment_by
                - quarters_to_decrement_by
                + quarter_based_export_site_source.default_batch_size_in_quarters
        )

        next_end_quarter_and_year = quarter_based_export_site_source._get_next_end_quarter_and_year(
            quarters_to_increment_by=quarters_to_increment_by, quarters_to_decrement_by=quarters_to_decrement_by)

        expected_next_end_quarter_and_year = update_quarter_year_tuple(
            start_quarter_and_year, quarters_to_increment_by=net_quarters_to_increment_by,
            quarters_to_decrement_by=0)
        self.assertTupleEqual(next_end_quarter_and_year, expected_next_end_quarter_and_year)

    def test_get_next_end_quarter_and_year_with_initial_end_quarter_and_year(self):
        """
        Should get the next end_quarter_and_year as end_quarter_and_year plus
        net quarter increment when end_quarter_and_year is given,
        regardless of start_quarter_and_year and default_batch_size_in_quarters
        """
        end_quarter_and_year = (1, 2020)
        start_quarter_and_year = (3, 2020,)
        quarters_to_increment_by = 4
        quarters_to_decrement_by = 18
        net_quarters_to_increment_by = quarters_to_increment_by - quarters_to_decrement_by

        quarter_based_export_site_source = ChildQuarterBasedExportSiteSource()
        quarter_based_export_site_source.end_quarter_and_year = end_quarter_and_year
        quarter_based_export_site_source.start_quarter_and_year = start_quarter_and_year

        next_end_quarter_and_year = quarter_based_export_site_source._get_next_end_quarter_and_year(
            quarters_to_increment_by=quarters_to_increment_by, quarters_to_decrement_by=quarters_to_decrement_by)

        expected_next_end_quarter_and_year = update_quarter_year_tuple(
            end_quarter_and_year, quarters_to_increment_by=net_quarters_to_increment_by,
            quarters_to_decrement_by=0)
        self.assertTupleEqual(next_end_quarter_and_year, expected_next_end_quarter_and_year)


if __name__ == '__main__':
    main()



"""Module containing a base source class"""

from datetime import date
from typing import Optional, List, Dict, Any, Iterator

from pydantic import BaseModel


class BaseSource(BaseModel):
    base_uri: Optional[str] = None
    name: Optional[str] = None
    attributes: List[str] = None

    def get(self) -> Iterator[Dict[str, Any]]:
        """Returns data from a given date to a given date as an iterator"""
        yield from self._query_data_source()

    def _query_data_source(self, *args, **kwargs) -> Iterator[Dict[str, Any]]:
        """Queries a given start and end date and returns an iterator with data records"""
        raise NotImplementedError(
            'Implement the _query_data_source method to return an iterator of dictionaries')



"""Module containing configurations for the date-based database data source"""

from datetime import date
from typing import Dict, Any, Optional, Type, Generator, Iterator

from sqlalchemy import text

from .config import DatabaseSourceDataModel
from ..base.date_based import DateBasedBaseSource
from ...destinations.database.config import DatabaseConnectionConfig
from ...destinations.database.connection import DatabaseConnection


class DateBasedDatabaseSource(DateBasedBaseSource):
    """Base class for the database source that depends on dates"""
    db_configuration: DatabaseConnectionConfig
    sql_query: str
    batch_size: int = 200
    data_model_class: Type[DatabaseSourceDataModel]

    def _get_params(self, start_date: Optional[date] = None, end_date: Optional[date] = None) -> Dict[Any, Any]:
        """An overridable method to return params dict to use in the SQL execute"""
        return {}

    def _query_data_source(self, start_date: date, end_date: date) -> Iterator[Dict[str, Any]]:
        """Queries a given start and end date and returns an iterator with data records"""
        params = self._get_params(start_date=start_date, end_date=end_date)

        with DatabaseConnection.get_db_connection(db_connection_config=self.db_configuration) as db_connection:
            result_proxy = db_connection.connection_engine.execution_options(stream_results=True).execute(
                text(self.sql_query), **params)

            while 'batch is not empty':
                # for memory efficiency, I have decided to query in batches of default size 200
                batch = result_proxy.fetchmany(size=self.batch_size)

                if not batch:
                    break

                yield from [self.data_model_class.from_orm(record).dict() for record in batch]

            result_proxy.close()

        yield from []



"""Configurations for the database package"""
from pydantic import BaseModel


class DatabaseConnectionConfig(BaseModel):
    db_uri: str
    pool_size: int = 5
    reflect: bool = True
    autocommit: bool = False
    pool_timeout: int = 60



"""Module With utilities for testing"""

import signal


class TestingTimeoutError(Exception):
    """Exception to be raised when a test has gone past its timeout"""
    pass


class TimeoutManager:
    """
    This provides a context manager to wrap around code
    that is to timeout if it takes longer than a given amount of time
    """

    def __init__(self, seconds, error_message=None):
        if error_message is None:
            error_message = f'Timeout after {seconds}s.'
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise TestingTimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.alarm(0)



"""Module containing tests for the logging utility functions"""

import logging
import os
from sys import getsizeof
from unittest import TestCase, main
from judah.utils.logging import setup_rotating_file_logger


_PARENT_FOLDER = os.path.dirname(__file__)
_MOCK_ASSET_FOLDER_PATH = os.path.join(os.path.dirname(_PARENT_FOLDER), 'assets')
_TEXT_OF_5000_BYTES = """\
Lorem ipsum dolor sit amet, consectetur adipiscing elit. Proin maximus, justo placerat tristique blandit, metus quam semper est, ac posuere metus lacus id ante. Orci varius natoque penatibus et magnis dis parturient montes, nascetur ridiculus mus. Morbi nec eros in lorem accumsan hendrerit. In varius risus in lectus interdum ornare. Quisque ut massa elit. Vestibulum sit amet neque lorem. Integer porttitor est nec euismod placerat.

Ut venenatis blandit tincidunt. Vestibulum bibendum orci purus, auctor posuere quam gravida ac. Vestibulum vehicula, ante ac molestie sagittis, sem sapien varius ante, vitae lacinia tortor tortor eget dui. Pellentesque commodo tempor tempus. Curabitur fringilla maximus massa id eleifend. Ut maximus felis ut faucibus lacinia. Phasellus hendrerit velit lorem, in varius nisl pretium et. Sed mattis turpis dolor, non finibus orci porta ut. Lorem ipsum dolor sit amet, consectetur adipiscing elit. Maecenas ullamcorper malesuada facilisis. In sem turpis, aliquam ac orci in, vehicula ullamcorper sem. Etiam nec blandit sem. Praesent sit amet pharetra felis.

Aliquam erat volutpat. Donec non magna a dolor convallis accumsan. Nulla efficitur a lacus in sollicitudin. Nullam ac pellentesque metus, sed aliquam sapien. Suspendisse eu ex vel augue vulputate pharetra. Quisque eu rhoncus augue. Nullam pulvinar nulla lacus, eu finibus magna mattis vel. Donec convallis a sapien non maximus. Donec sollicitudin tempus leo, vel varius libero rhoncus non. Duis porttitor orci et felis bibendum, at pharetra elit vulputate. In hac habitasse platea dictumst. Sed dolor odio, rhoncus nec eleifend eget, consectetur vehicula ante. Vestibulum dignissim sit amet felis eu pellentesque.

Vivamus purus tortor, lobortis vitae sodales quis, auctor non diam. Donec rutrum risus vitae nisi tristique auctor. Donec tempor sed eros sed ultrices. Maecenas id congue ipsum. Etiam blandit tincidunt magna, a commodo mi tristique sit amet. Proin vulputate tempus aliquet. Donec ultricies ipsum et sodales molestie. Aenean dignissim diam nisl, a eleifend nulla pharetra accumsan. Curabitur scelerisque quam sed dolor iaculis, vitae dapibus turpis accumsan. Ut quis leo lectus. Nunc at suscipit ante. Duis non consequat turpis. Mauris nisi nibh, pellentesque id neque sed, ultrices congue urna. Proin blandit nec risus eu lobortis.

Vestibulum aliquet lectus elit, id vestibulum felis convallis at. Suspendisse semper elit sed leo ultrices, et lacinia orci vestibulum. Nam ullamcorper pharetra nulla, gravida efficitur urna rhoncus in. Praesent ultricies urna vel pretium interdum. Vivamus luctus ultrices dignissim. Ut vel mollis lectus. In rhoncus tortor sed nisi tincidunt, ac luctus dolor gravida.

Cras convallis sapien nisl, a ultrices ex porta nec. In vitae lectus ut nisi consectetur tempus. Duis vel lacus eget tellus feugiat vestibulum in vel tellus. Nulla eget turpis neque. Aliquam finibus rhoncus pretium. Aliquam ac mi eu augue sagittis pellentesque at ac quam. Proin at augue id neque imperdiet rutrum. Ut egestas eleifend massa ut egestas. Vestibulum malesuada eros eget gravida fermentum. Ut in luctus ante. Curabitur sodales feugiat risus. Etiam pulvinar bibendum ornare. Suspendisse gravida vel nisi vitae volutpat. Mauris aliquam, risus quis placerat accumsan, eros neque euismod eros, id vestibulum purus nisl a velit.

Donec non enim est. Nulla diam metus, ultricies eget dui vel, feugiat ultricies magna. Nunc efficitur leo sed pulvinar mattis. Vestibulum dictum odio sed sapien fermentum finibus. Praesent aliquam hendrerit quam vitae eleifend. Ut bibendum fringilla ligula, non rutrum purus. Donec in libero cursus, elementum est eu, finibus dolor. Cras semper iaculis dignissim. Donec elementum mi ante, at tempus nisi accumsan in. In placerat, tellus ut imperdiet sagittis, erat mauris imperdiet nunc, quis convallis nisi lectus sit amet odio.

Sed tincidunt ut erat eu tempus. Nullam id est sit amet velit gravida mollis. Sed iaculis mi nec velit efficitur molestie. Ut sed ligula libero. Fusce neque erat, malesuada vel rutrum at, mattis id augue. Integer vel diam a enim congue ultrices ac at mauris. Morbi vestibulum porttitor massa eu sagittis. Pellentesque habitant morbi tristique senectus et netus et malesuada fames ac turpis egestas.

Mauris commodo accumsan convallis. Donec rutrum aliquam nulla, eget pretium sem varius quis. Aenean nec sem vestibulum, tristique tortor et, semper arcu. Nulla facilisi. In hac habitasse platea dictumst. Etiam a ligula eu felis semper malesuada non in nulla. Pellentesque habitant morbi tristique senectus et netus et malesuada fames ac turpis egestas. Curabitur faucibus tincidunt mattis. Etiam id nunc leo. Aenean imperdiet risus sit amet nunc semper posuere. Etiam in nibh id felis accumsan porta nec quis nibh. Nam mollis tortor vitae turpis pellentesque rutrum. Mauris porta turpis et orci volutpat auctor. Quisque quis interdum velit. Pellentesque ut nisl non nf..
"""


class TestLoggingUtilities(TestCase):
    """Tests for the logging utility functions"""

    def setUp(self) -> None:
        """Initialize some common variables"""
        self.log_file_path = os.path.join(_MOCK_ASSET_FOLDER_PATH, 'error.log')
        self.log_backup_file_path = os.path.join(_MOCK_ASSET_FOLDER_PATH, 'error.log.1')

    def test_setup_rotating_file_logger(self):
        """Makes the logger to use a file that rotates when a given size is reached"""
        logger = logging.getLogger('random-test')

        max_bytes = getsizeof(_TEXT_OF_5000_BYTES)
        self.assertEqual(max_bytes, 5000)

        setup_rotating_file_logger(logger=logger, file_path=self.log_file_path, max_bytes=max_bytes)

        self.assertFalse(os.path.isfile(self.log_backup_file_path))

        logger.error(_TEXT_OF_5000_BYTES)
        logger.error(_TEXT_OF_5000_BYTES)
        self.assertTrue(os.path.isfile(self.log_backup_file_path))

        file_size_of_log_backup = os.path.getsize(self.log_backup_file_path)
        file_size_of_log = os.path.getsize(self.log_file_path)
        max_possible_size = 5000 + 1

        self.assertLessEqual(file_size_of_log, max_possible_size)
        self.assertLessEqual(file_size_of_log_backup, max_possible_size)

    def tearDown(self) -> None:
        """Clean up"""
        if os.path.isfile(self.log_file_path):
            os.remove(self.log_file_path)

        if os.path.isfile(self.log_backup_file_path):
            os.remove(self.log_backup_file_path)


if __name__ == '__main__':
    main()



"""Module containing the controller class for getting data from the date based export site to the database"""

from datetime import date
from typing import Type, Optional

from ..base.date_based import DateBasedBaseController
from ...destinations.database.model import DatabaseBaseModel
from ...sources.export_site.date_based import DateBasedExportSiteSource
from ...utils.dates import get_default_historical_start_date


class DateBasedExportSiteToDatabaseController(DateBasedBaseController):
    """
    Class for the controller that downloads data from the date-based export site
    and saves it in a database
    """
    _source: DateBasedExportSiteSource

    destination_model_class: Type[DatabaseBaseModel]
    source_class: Type[DateBasedExportSiteSource]
    start_date: Optional[date] = get_default_historical_start_date()



"""Module containing tests for the assets utility functions"""
import os
import shutil
from collections import Iterator
from datetime import datetime, timedelta
from unittest import TestCase, main
from unittest.mock import Mock, patch
from judah.utils.assets import get_asset_path, get_timestamped_folder_name, get_timestamped_folder, \
    get_csv_download_location, get_xml_download_location, read_xls_file, read_csv_file, read_file, delete_parent_folder, \
    FileType, FileOptions

_ROOT_PATH = os.getcwd()
_PARENT_FOLDER = os.path.dirname(__file__)
_MOCK_ASSET_FOLDER_PATH = os.path.join(os.path.dirname(_PARENT_FOLDER), 'assets')


class TestAssetsUtilities(TestCase):
    """Tests for the assets utility functions"""

    def setUp(self) -> None:
        """Initialize a few variables"""
        self.default_time_format = '%m_%d_%Y_%H_%M_%S'

    def test_get_asset_path_default(self):
        """returns the asset folder path at the root of where the script is run when no parameter is passed"""
        expected_asset_folder_path = os.path.join(_ROOT_PATH, 'assets')
        self.assertEqual(get_asset_path(), expected_asset_folder_path)

    def test_get_asset_path(self):
        """returns the path to a given folder in the asset folder"""
        asset_name = 'photos'
        expected_asset_folder_path = os.path.join(_ROOT_PATH, 'assets', asset_name)
        self.assertEqual(get_asset_path(asset_name=asset_name), expected_asset_folder_path)

    def test_get_timestamped_folder_name(self):
        """returns a folder name with a timestamp in it name"""
        time_now = datetime.now()
        default_folder_name = get_timestamped_folder_name()
        default_folder_timestamp = datetime.strptime(default_folder_name, self.default_time_format)
        self.assertLessEqual(default_folder_timestamp - time_now, timedelta(minutes=1))

    def test_get_timestamped_folder_name_with_suffix(self):
        """returns a folder name with a timestamp in it name and the suffix"""
        time_now = datetime.now()
        suffix = 'hey'
        suffixed_folder_name = get_timestamped_folder_name(suffix=suffix)
        timestamp_string = suffixed_folder_name.split(suffix)[0]
        suffixed_folder_timestamp = datetime.strptime(timestamp_string, self.default_time_format)
        self.assertLessEqual(suffixed_folder_timestamp - time_now, timedelta(minutes=1))
        self.assertEqual(suffixed_folder_name.index(suffix), len(suffixed_folder_name) - len(suffix))

    def test_get_timestamped_folder_name_with_prefix(self):
        """returns a folder name with a timestamp in it name and the prefix"""
        time_now = datetime.now()
        prefix = 'hey'
        prefixed_folder_name = get_timestamped_folder_name(prefix=prefix)
        timestamp_string = prefixed_folder_name.split(prefix)[1]
        prefixed_folder_timestamp = datetime.strptime(timestamp_string, self.default_time_format)

        self.assertLessEqual(prefixed_folder_timestamp - time_now, timedelta(minutes=1))
        self.assertEqual(prefixed_folder_name.index(prefix), 0)

    def test_get_timestamped_folder_name_with_other_format(self):
        """returns a folder name with a timestamp in it with a different time format"""
        other_format = '%Y-%m'
        time_now = datetime.now()
        other_format_folder_name = get_timestamped_folder_name(date_format=other_format)
        default_folder_timestamp = datetime.strptime(other_format_folder_name, other_format)
        self.assertLessEqual(default_folder_timestamp - time_now, timedelta(minutes=1))

    def test_get_timestamped_folder(self):
        """returns a folder path with a timestamp in it name and creates the folder if it does not exist"""
        time_now = datetime.now()
        default_folder_path = get_timestamped_folder(root_path=_ROOT_PATH)
        default_folder_name = os.path.basename(default_folder_path)
        default_folder_timestamp = datetime.strptime(default_folder_name, self.default_time_format)

        self.assertLessEqual(default_folder_timestamp - time_now, timedelta(minutes=1))
        self.assertEqual(os.path.dirname(default_folder_path), _ROOT_PATH)

        self.assertTrue(os.path.isdir(default_folder_path))
        shutil.rmtree(default_folder_path)

    def test_get_timestamped_folder_with_suffix(self):
        """returns a folder path with a timestamp in it name and the suffix and creates the folder if not exist"""
        time_now = datetime.now()
        suffix = 'hey'
        suffixed_folder_path = get_timestamped_folder(suffix=suffix, root_path=_ROOT_PATH)
        suffixed_folder_name = os.path.basename(suffixed_folder_path)
        timestamp_string = suffixed_folder_name.split(suffix)[0]
        suffixed_folder_timestamp = datetime.strptime(timestamp_string, self.default_time_format)

        self.assertLessEqual(suffixed_folder_timestamp - time_now, timedelta(minutes=1))
        self.assertEqual(suffixed_folder_name.index(suffix), len(suffixed_folder_name) - len(suffix))
        self.assertEqual(os.path.dirname(suffixed_folder_path), _ROOT_PATH)

        self.assertTrue(os.path.isdir(suffixed_folder_path))
        shutil.rmtree(suffixed_folder_path)

    def test_get_timestamped_folder_with_prefix(self):
        """returns a folder path with a timestamp in it name and the prefix"""
        time_now = datetime.now()
        prefix = 'hey'
        prefixed_folder_path = get_timestamped_folder(prefix=prefix, root_path=_ROOT_PATH)
        prefixed_folder_name = os.path.basename(prefixed_folder_path)
        timestamp_string = prefixed_folder_name.split(prefix)[1]
        prefixed_folder_timestamp = datetime.strptime(timestamp_string, self.default_time_format)

        self.assertLessEqual(prefixed_folder_timestamp - time_now, timedelta(minutes=1))
        self.assertEqual(prefixed_folder_name.index(prefix), 0)
        self.assertEqual(os.path.dirname(prefixed_folder_path), _ROOT_PATH)

        self.assertTrue(os.path.isdir(prefixed_folder_path))
        shutil.rmtree(prefixed_folder_path)

    def test_get_timestamped_folder_with_other_format(self):
        """returns a folder path with a timestamp in it with a different time format"""
        other_format = '%Y-%m'
        time_now = datetime.now()
        other_format_folder_path = get_timestamped_folder(date_format=other_format, root_path=_ROOT_PATH)
        other_format_folder_name = os.path.basename(other_format_folder_path)
        folder_timestamp = datetime.strptime(other_format_folder_name, other_format)

        self.assertLessEqual(folder_timestamp - time_now, timedelta(minutes=1))
        self.assertTrue(os.path.isdir(other_format_folder_path))
        shutil.rmtree(other_format_folder_path)

    def test_get_csv_download_location(self):
        """
        returns the path to a timestamped folder prefixed by the dataset name
        within the csv folder in the assets folder on the root of the project
        """
        dataset_name = 'tmp'
        time_now = datetime.now()
        expected_csv_dataset_folder_path = os.path.join(_ROOT_PATH, 'assets', 'csv')
        downloaded_csv_folder_path = get_csv_download_location(dataset_name=dataset_name)

        parent_folder_of_downloaded_csv_folder = os.path.dirname(downloaded_csv_folder_path)
        downloaded_csv_folder_name = os.path.basename(downloaded_csv_folder_path)
        downloaded_csv_folder_timestamp = datetime.strptime(downloaded_csv_folder_name.strip(dataset_name),
                                                            self.default_time_format)

        self.assertEqual(parent_folder_of_downloaded_csv_folder, expected_csv_dataset_folder_path)
        self.assertLessEqual(downloaded_csv_folder_timestamp - time_now, timedelta(minutes=1))

        self.assertTrue(os.path.isdir(downloaded_csv_folder_path))
        shutil.rmtree(downloaded_csv_folder_path)

    def test_get_xml_download_location(self):
        """
        returns the path to a timestamped folder prefixed by the dataset name
        within the xml folder in the assets folder on the root of the project
        """
        dataset_name = 'tmp'
        time_now = datetime.now()
        expected_xml_dataset_folder_path = os.path.join(_ROOT_PATH, 'assets', 'xml')
        downloaded_xml_folder_path = get_xml_download_location(dataset_name=dataset_name)

        parent_folder_of_downloaded_xml_folder = os.path.dirname(downloaded_xml_folder_path)
        downloaded_xml_folder_name = os.path.basename(downloaded_xml_folder_path)
        downloaded_xml_folder_timestamp = datetime.strptime(downloaded_xml_folder_name.strip(dataset_name),
                                                            self.default_time_format)

        self.assertEqual(parent_folder_of_downloaded_xml_folder, expected_xml_dataset_folder_path)
        self.assertLessEqual(downloaded_xml_folder_timestamp - time_now, timedelta(minutes=1))

        self.assertTrue(os.path.isdir(downloaded_xml_folder_path))
        shutil.rmtree(downloaded_xml_folder_path)

    def test_read_csv_file(self):
        """Reads a CSV file returning dict by dict in an iterator"""
        mock_csv_file_path = os.path.join(_MOCK_ASSET_FOLDER_PATH, 'mock.csv')
        csv_iterator = read_csv_file(mock_csv_file_path)

        expected_data = [
            {"Date": "09/03/2020", "number": "1", "period_from": "00:00", "period_until": "00:15", "Capacity": "16616"},
            {"Date": "09/03/2020", "number": "2", "period_from": "00:15", "period_until": "00:30", "Capacity": "16616"},
            {"Date": "09/03/2020", "number": "3", "period_from": "00:30", "period_until": "00:45", "Capacity": "16616"},
            {"Date": "09/03/2020", "number": "4", "period_from": "00:45", "period_until": "01:00", "Capacity": "16620"},
        ]

        self.assertIsInstance(csv_iterator, Iterator)

        data = [record for record in csv_iterator]
        self.assertListEqual(data, expected_data)

    def test_read_xls_file(self):
        """Reads an XLS file returning dict by dict in an iterator"""
        mock_xls_file_path = os.path.join(_MOCK_ASSET_FOLDER_PATH, 'mock.xls')
        xls_iterator = read_xls_file(mock_xls_file_path, xls_sheet_index=0)

        expected_data = [
            {"number": 1.0, "power": 16616.0},
            {"number": 2.0, "power": 16616.0},
            {"number": 3.0, "power": 16616.0},
            {"number": 4.0, "power": 16616.0},
        ]

        self.assertIsInstance(xls_iterator, Iterator)

        data = [record for record in xls_iterator]
        self.assertListEqual(data, expected_data)

    @patch('judah.utils.assets.read_xml_file')
    def test_read_file_when_xml(self, mock_read_xml_file: Mock):
        """Reads a file as XML if file_type is XML"""
        mock_xml_file_path = os.path.join(_MOCK_ASSET_FOLDER_PATH, 'mock.xml')
        options = FileOptions()

        xml_iterator = read_file(mock_xml_file_path, file_type=FileType.XML, options=options)
        data = [record for record in xml_iterator]

        mock_read_xml_file.assert_called_once_with(file_path=mock_xml_file_path, to_dict=True,
                                                   records_tag=options.xml_records_tag)

    @patch('judah.utils.assets.read_xls_file')
    def test_read_file_when_xls(self, mock_read_xls_file: Mock):
        """Reads a file as XLS if file_type is XLS"""
        mock_xls_file_path = os.path.join(_MOCK_ASSET_FOLDER_PATH, 'mock.xls')
        options = FileOptions()

        xls_iterator = read_file(mock_xls_file_path, file_type=FileType.XLS, options=options)
        data = [record for record in xls_iterator]

        mock_read_xls_file.assert_called_once_with(file_path=mock_xls_file_path, **dict(options))

    @patch('judah.utils.assets.read_csv_file')
    def test_read_file_when_csv(self, mock_read_csv_file: Mock):
        """Reads a file as CSV if file_type is CSV"""
        mock_csv_file_path = os.path.join(_MOCK_ASSET_FOLDER_PATH, 'mock.csv')
        options = FileOptions()

        csv_iterator = read_file(mock_csv_file_path, file_type=FileType.CSV, options=options)
        data = [record for record in csv_iterator]

        mock_read_csv_file.assert_called_once_with(file_path=mock_csv_file_path, **dict(options))

    def test_delete_parent_folder(self):
        """Deletes the parent folder of a given file"""
        dummy_folder_path = os.path.join(_MOCK_ASSET_FOLDER_PATH, 'dummy')
        dummy_file_path = os.path.join(dummy_folder_path, 'dummy.csv')

        if not os.path.exists(dummy_folder_path):
            os.makedirs(dummy_folder_path)

        file_holder = open(dummy_file_path, 'w+')
        file_holder.close()

        delete_parent_folder(dummy_file_path)
        self.assertFalse(os.path.exists(dummy_folder_path))


if __name__ == '__main__':
    main()



"""Module containing tests for the date based Exports Site data source"""
import os
from collections import Iterator
from datetime import date, timedelta
from typing import Optional
from unittest import TestCase, main
from unittest.mock import patch, Mock, call

from selenium import webdriver

from judah.sources.export_site.date_based import DateBasedExportSiteSource

_PARENT_FOLDER = os.path.dirname(__file__)
_MOCK_ASSET_FOLDER_PATH = os.path.join(os.path.dirname(os.path.dirname(_PARENT_FOLDER)), 'assets')


class ChildDateBasedExportSiteSource(DateBasedExportSiteSource):
    """Child export site source that just picks a file from the file system"""
    base_uri: str = 'http://example.com'
    name: str = 'test_export_site_source'

    def _download_file(self, start_date: date, end_date: date) -> Optional[str]:
        """Downloads the CSV from the export site and returns the path to it"""
        return None


class TestDateBasedExportSiteSource(TestCase):
    """Tests for the DateBasedExportSiteSource"""

    def setUp(self) -> None:
        """Initialize some variables"""
        self.mock_csv_file_path = os.path.join(_MOCK_ASSET_FOLDER_PATH, 'mock.csv')
        self.expected_data = [
            {"Date": "09/03/2020", "number": "1", "period_from": "00:00", "period_until": "00:15", "Capacity": "16616"},
            {"Date": "09/03/2020", "number": "2", "period_from": "00:15", "period_until": "00:30", "Capacity": "16616"},
            {"Date": "09/03/2020", "number": "3", "period_from": "00:30", "period_until": "00:45", "Capacity": "16616"},
            {"Date": "09/03/2020", "number": "4", "period_from": "00:45", "period_until": "01:00", "Capacity": "16620"},
        ]

    @patch('judah.sources.export_site.date_based.visit_website')
    def test_initialize_chrome(self, mock_visit_website):
        """
        Should initialize Chrome in case it is not yet initialized and visits the base url
        """
        date_based_export_site_source = ChildDateBasedExportSiteSource()
        self.assertIsNone(date_based_export_site_source.chrome)

        date_based_export_site_source._initialize_chrome()

        self.assertIsInstance(date_based_export_site_source.chrome, webdriver.Chrome)
        mock_visit_website.assert_called_once_with(
            driver=date_based_export_site_source.chrome, website_url=date_based_export_site_source.base_uri)

        date_based_export_site_source.chrome.close()
        date_based_export_site_source.chrome.quit()

    @patch('judah.sources.export_site.date_based.delete_parent_folder')
    @patch.object(ChildDateBasedExportSiteSource, '_initialize_chrome')
    @patch.object(ChildDateBasedExportSiteSource, '_download_file')
    def test_query_data_source(self, mock_download_file, mock_initialize_chrome, mock_delete_parent_folder):
        """
        Should query a given start and end date and return an iterator with data records and then deletes folder
        """
        # initializations
        mock_download_file.return_value = self.mock_csv_file_path
        date_based_export_site_source = ChildDateBasedExportSiteSource()
        start_date = date(year=2020, month=3, day=9)
        end_date = date(year=2020, month=3, day=13)

        # method call
        response = date_based_export_site_source._query_data_source(start_date=start_date, end_date=end_date)
        self.assertIsInstance(response, Iterator)

        data = [record for record in response]

        # assertions
        mock_initialize_chrome.assert_called_once()
        mock_download_file.assert_called_once_with(start_date=start_date, end_date=end_date)
        mock_delete_parent_folder.assert_called_once_with(self.mock_csv_file_path)
        self.assertListEqual(data, self.expected_data)

    @patch('judah.sources.export_site.date_based.delete_parent_folder')
    @patch.object(ChildDateBasedExportSiteSource, '_initialize_chrome')
    @patch.object(ChildDateBasedExportSiteSource, '_download_file')
    def test_query_data_source_no_file_downloaded(self, mock_download_file, mock_initialize_chrome,
                                                  mock_delete_parent_folder):
        """
        Should query a given start and end date and return an empty iterator if there is no file downloaded
        """
        # initializations
        mock_download_file.return_value = None
        date_based_export_site_source = ChildDateBasedExportSiteSource()
        start_date = date(year=2020, month=3, day=9)
        end_date = date(year=2020, month=3, day=13)

        # method call
        response = date_based_export_site_source._query_data_source(start_date=start_date, end_date=end_date)
        self.assertIsInstance(response, Iterator)

        data = [record for record in response]

        # assertions
        mock_initialize_chrome.assert_called_once()
        mock_download_file.assert_called_once_with(start_date=start_date, end_date=end_date)
        mock_delete_parent_folder.assert_not_called()
        self.assertListEqual(data, [])

    @patch.object(ChildDateBasedExportSiteSource, '_initialize_chrome')
    def test_del(self, mock_initialize_chrome):
        """
        Should quit chrome on deletion
        """
        date_based_export_site_source = ChildDateBasedExportSiteSource()
        date_based_export_site_source.chrome = Mock(spec=webdriver.Chrome)
        date_based_export_site_source.__del__()
        date_based_export_site_source.chrome.quit.assert_called_once()

    @patch.object(ChildDateBasedExportSiteSource, '_query_data_source')
    @patch.object(ChildDateBasedExportSiteSource, '_get_next_end_date')
    @patch.object(ChildDateBasedExportSiteSource, '_get_next_start_date')
    def test_get(self, mock_get_next_start_date, mock_get_next_end_date, mock_query_data_source):
        """
        Should return data from a given date to a given date as an iterator
        """
        start_date = date(year=2020, month=3, day=9)
        end_date = date(year=2020, month=3, day=13)
        mock_get_next_start_date.return_value = start_date
        mock_get_next_end_date.return_value = end_date
        mock_query_data_source.return_value = (record for record in self.expected_data)

        date_based_export_site_source = ChildDateBasedExportSiteSource()
        response = date_based_export_site_source.get()

        self.assertIsInstance(response, Iterator)

        data = [record for record in response]

        self.assertListEqual(data, self.expected_data)
        mock_get_next_start_date.assert_has_calls(
            [call(), call(days_to_increment_by=date_based_export_site_source.default_batch_size_in_days)])
        mock_get_next_end_date.assert_called_once()

    @patch.object(ChildDateBasedExportSiteSource, '_query_data_source')
    @patch.object(ChildDateBasedExportSiteSource, '_get_next_end_date')
    @patch.object(ChildDateBasedExportSiteSource, '_get_next_start_date')
    def test_get_end_date_earlier_than_start_date(self, mock_get_next_start_date, mock_get_next_end_date,
                                                  mock_query_data_source):
        """
        Should return an iterator of an empty list if the end_date is earlier than start_date
        """
        start_date = date(year=2020, month=3, day=30)
        end_date = date(year=2020, month=3, day=13)
        mock_get_next_start_date.return_value = start_date
        mock_get_next_end_date.return_value = end_date

        date_based_export_site_source = ChildDateBasedExportSiteSource()
        response = date_based_export_site_source.get()

        self.assertIsInstance(response, Iterator)

        data = [record for record in response]

        self.assertListEqual(data, [])
        mock_get_next_start_date.assert_called_once()
        mock_get_next_end_date.assert_called_once()
        mock_query_data_source.assert_not_called()

    def test_get_next_start_date_with_no_initial_start_date_and_no_end_date(self):
        """
        Should get the next start date as today minus the default_batch_size_in_days plus net day increment
        when only default_batch_size_in_days is given
        """
        today = date.today()
        days_to_increment_by = 4
        days_to_decrement_by = 8
        net_days_to_increment_by = days_to_increment_by - days_to_decrement_by

        date_based_export_site_source = ChildDateBasedExportSiteSource()
        next_start_date = date_based_export_site_source._get_next_start_date(
            days_to_increment_by=days_to_increment_by, days_to_decrement_by=days_to_decrement_by)

        expected_next_start_date = today + timedelta(days=net_days_to_increment_by) - timedelta(
            days=date_based_export_site_source.default_batch_size_in_days)
        self.assertEqual(next_start_date, expected_next_start_date)

    def test_get_next_start_date_given_end_date(self):
        """
        Should get the next start date as end_date minus the default_batch_size_in_days plus net day increment
        when default_batch_size_in_days and end_date are given
        """
        end_date = date(year=2020, month=3, day=13)
        days_to_increment_by = 4
        days_to_decrement_by = 8
        net_days_to_increment_by = days_to_increment_by - days_to_decrement_by

        date_based_export_site_source = ChildDateBasedExportSiteSource()
        date_based_export_site_source.end_date = end_date

        next_start_date = date_based_export_site_source._get_next_start_date(
            days_to_increment_by=days_to_increment_by, days_to_decrement_by=days_to_decrement_by)

        expected_next_start_date = end_date + timedelta(days=net_days_to_increment_by) - timedelta(
            days=date_based_export_site_source.default_batch_size_in_days)
        self.assertEqual(next_start_date, expected_next_start_date)

    def test_get_next_start_date_with_initial_start_date(self):
        """
        Should get the next start date as start_date plus net day increment when start_date is given,
        regardless of end_date and default_batch_size_in_days
        """
        end_date = date(year=2020, month=3, day=13)
        start_date = date(year=2020, month=3, day=12)
        days_to_increment_by = 4
        days_to_decrement_by = 8
        net_days_to_increment_by = days_to_increment_by - days_to_decrement_by

        date_based_export_site_source = ChildDateBasedExportSiteSource()
        date_based_export_site_source.end_date = end_date
        date_based_export_site_source.start_date = start_date

        next_start_date = date_based_export_site_source._get_next_start_date(
            days_to_increment_by=days_to_increment_by, days_to_decrement_by=days_to_decrement_by)

        expected_next_start_date = start_date + timedelta(days=net_days_to_increment_by)
        self.assertEqual(next_start_date, expected_next_start_date)

    def test_get_next_end_date_with_no_initial_start_date_and_no_end_date(self):
        """
        Should get the next end date as today plus net day increment
        when only default_batch_size_in_days is given: Assumption is today is end date if no increment or decrement
        """
        today = date.today()
        days_to_increment_by = 4
        days_to_decrement_by = 9
        net_days_to_increment_by = days_to_increment_by - days_to_decrement_by

        date_based_export_site_source = ChildDateBasedExportSiteSource()
        next_end_date = date_based_export_site_source._get_next_end_date(
            days_to_increment_by=days_to_increment_by, days_to_decrement_by=days_to_decrement_by)

        expected_next_end_date = today + timedelta(days=net_days_to_increment_by)
        self.assertEqual(next_end_date, expected_next_end_date)

    def test_get_next_end_date_given_start_date(self):
        """
        Should get the next end date as start_date plus the net day increment plus default_batch_size_in_days
        when default_batch_size_in_days and start_date are given and no initial end_date is given
        """
        start_date = date(year=2020, month=3, day=13)
        days_to_increment_by = 4
        days_to_decrement_by = 9
        net_days_to_increment_by = days_to_increment_by - days_to_decrement_by

        date_based_export_site_source = ChildDateBasedExportSiteSource()
        date_based_export_site_source.start_date = start_date

        next_end_date = date_based_export_site_source._get_next_end_date(
            days_to_increment_by=days_to_increment_by, days_to_decrement_by=days_to_decrement_by)

        expected_next_end_date = start_date + timedelta(days=net_days_to_increment_by) + timedelta(
            days=date_based_export_site_source.default_batch_size_in_days)
        self.assertEqual(next_end_date, expected_next_end_date)

    def test_get_next_end_date_with_initial_end_date(self):
        """
        Should get the next end date as end_date plus net day increment when end_date is given,
        regardless of start_date and default_batch_size_in_days
        """
        end_date = date(year=2020, month=3, day=13)
        start_date = date(year=2020, month=3, day=12)
        days_to_increment_by = 4
        days_to_decrement_by = 18
        net_days_to_increment_by = days_to_increment_by - days_to_decrement_by

        date_based_export_site_source = ChildDateBasedExportSiteSource()
        date_based_export_site_source.end_date = end_date
        date_based_export_site_source.start_date = start_date

        next_end_date = date_based_export_site_source._get_next_end_date(
            days_to_increment_by=days_to_increment_by, days_to_decrement_by=days_to_decrement_by)

        expected_next_end_date = end_date + timedelta(days=net_days_to_increment_by)
        self.assertEqual(next_end_date, expected_next_end_date)


if __name__ == '__main__':
    main()



"""Module containing the export site data source"""
from typing import Iterator, Dict, Any, Optional

from selenium import webdriver

from ..base import BaseSource
from ...utils.assets import read_file, delete_parent_folder, FileType, FileOptions, get_csv_download_location, \
    get_xml_download_location, get_asset_path
from ...utils.selenium import get_web_driver, WebDriverOptions, visit_website


class ExportSiteSource(BaseSource):
    default_file_name: str = 'export.csv'
    file_type: FileType = FileType.CSV
    file_options: FileOptions = FileOptions()

    chrome: Optional[webdriver.Chrome] = None
    download_folder_path: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True

    def _initialize_chrome(self):
        """Initializes Chrome in case it is not yet initialized"""
        if isinstance(self.chrome, webdriver.Chrome):
            return

        if self.file_type == FileType.CSV:
            self.download_folder_path = get_csv_download_location(dataset_name=self.name.replace(' ', '_'))
        elif self.file_type == FileType.XML:
            self.download_folder_path = get_xml_download_location(dataset_name=self.name.replace(' ', '_'))
        else:
            self.download_folder_path = get_asset_path()

        self.chrome = get_web_driver(
            WebDriverOptions(downloads_folder_location=self.download_folder_path))

        visit_website(driver=self.chrome, website_url=self.base_uri)

    def _download_file(self, *args, **kwargs) -> str:
        """Downloads the CSV from the export site and returns the path to it"""
        raise NotImplementedError('_download_file method should be implemented')

    def _query_data_source(self, *args, **kwargs) -> Iterator[Dict[str, Any]]:
        """Queries a given start and end date and returns an iterator with data records"""
        self._initialize_chrome()
        file_path = self._download_file(*args, **kwargs)

        if file_path is None:
            yield from []
        else:
            yield from read_file(
                file_path=file_path, file_type=self.file_type, options=self.file_options)
            delete_parent_folder(file_path)

    def __del__(self):
        """Clean up"""
        self.chrome.quit()



"""Module contains the class that transforms data by formatting the date"""
from datetime import datetime
from typing import Dict, Any, List, Union

from .base import BaseTransformer
from ..utils.dates import change_datetime_format


class DatetimeTransformer(BaseTransformer):
    """Transforms the format of the date field in a given data dictionary"""
    datetime_field: str = 'Date'
    source_datetime_format: str = '%m/%d/%Y'
    destination_datetime_format: str = '%Y-%m-%d'

    @classmethod
    def run(cls, data: Dict[Any, Any]):
        """Transforms the dates to the right format"""
        data_copy = data.copy()
        date_value = data_copy.get(cls.datetime_field, None)

        if date_value is not None:
            data_copy[cls.datetime_field] = change_datetime_format(
                datetime_value=date_value, old=cls.source_datetime_format,
                new=cls.destination_datetime_format)

        return data_copy


class MultipleFormatDatetimeTransformer(BaseTransformer):
    """
    Transforms the format of the datetime field in a given data dictionary
    from a number of possible source formats
    """
    datetime_field: str = 'Date'
    source_datetime_formats: List[str] = ['%m/%d/%Y']
    destination_datetime_format: str = '%Y-%m-%d'

    @classmethod
    def __format_datetime(cls, datetime_value: str) -> Union[str, datetime]:
        """Attempts to format the value but using the many possible source datetime formats"""
        new_value = datetime_value

        for old_datetime_format in cls.source_datetime_formats:
            try:
                new_value = change_datetime_format(
                    datetime_value=datetime_value, old=old_datetime_format,
                    new=cls.destination_datetime_format)
                break
            except ValueError:
                pass

        return new_value

    @classmethod
    def run(cls, data: Dict[Any, Any]):
        """Transforms the dates to the right format"""
        data_copy = data.copy()
        datetime_value = data_copy.get(cls.datetime_field, None)

        if datetime_value is not None:
            data_copy[cls.datetime_field] = cls.__format_datetime(datetime_value=datetime_value)

        return data_copy



"""A few configuration classes for the database source"""
from pydantic import BaseModel


class DatabaseSourceDataModel(BaseModel):
    """The base model for data models of the Database source"""

    class Config:
        orm_mode = True



"""Module with tests for the date based REST API source"""
from datetime import date, timedelta
from typing import Optional, Dict, Any
from unittest import TestCase, main
from unittest.mock import patch, call

import requests
from pydantic import BaseModel

from judah.sources.rest_api.date_based import DateBasedRestAPISource


class MockResponse(BaseModel):
    ok: bool = True
    status_code: int = 200
    mock_data: Optional[Dict[Any, Any]] = None

    def json(self, *args, **kwargs):
        """Dummy json method"""
        return self.mock_data


class TestDateBasedRestApiSource(TestCase):
    """Tests for the DateBasedRestApiSource"""

    def setUp(self) -> None:
        """Initialize some variables"""
        self.options = dict(response_data_key='results',
                            default_headers={
                                "Host": "judah.github.com"
                            },
                            start_date_param_name="start_date",
                            end_date_param_name="end_date",
                            date_format='%Y-%m-%d',
                            end_date=date.today(),
                            start_date=date.today() - timedelta(days=1))

    def test_get_with_params(self):
        """Makes a GET request with the params from the _get_params method"""
        mock_data = {'results': [{'foo': 'bar'}, {'foo2': 'bar2'}]}
        dummy_params = {'foo': 'bar', 'sample': 6}

        class DateBasedRestApiSourceWithParams(DateBasedRestAPISource):
            base_uri: Optional[str] = 'http://example.com'
            name: str = 'sample'

            def _get_params(self, **kwargs) -> Dict[Any, Any]:
                return dummy_params

        with patch.object(requests, 'get',
                          return_value=MockResponse(mock_data=mock_data)) as mock_get:
            source = DateBasedRestApiSourceWithParams(**self.options)
            data = list(source.get())

            url = f'{source.base_uri}/{source.name}'
            mock_get.assert_called_with(url=url, headers=self.options['default_headers'], params=dummy_params)

            self.assertListEqual(data, mock_data['results'])

    def test_get_with_default_params(self):
        """Makes a GET request with the default params of start_date and end_date"""
        mock_data = {'results': [{'foo': 'bar'}, {'foo2': 'bar2'}]}
        default_params = {
            self.options['start_date_param_name']: self.options['start_date'].strftime(self.options['date_format']),
            self.options['end_date_param_name']: self.options['end_date'].strftime(self.options['date_format'])}

        class DateBasedRestApiSourceWithDefaultParams(DateBasedRestAPISource):
            base_uri: Optional[str] = 'http://example.com'
            name: str = 'sample'

        with patch.object(requests, 'get',
                          return_value=MockResponse(mock_data=mock_data)) as mock_get:
            source = DateBasedRestApiSourceWithDefaultParams(**self.options)
            data = list(source.get())

            url = f'{source.base_uri}/{source.name}'
            mock_get.assert_called_with(url=url, headers=self.options['default_headers'], params=default_params)

            self.assertListEqual(data, mock_data['results'])

    def test_get_with_headers(self):
        """Makes a GET request with the headers from the _get_headers method"""
        mock_data = {'results': [{'foo': 'bar'}, {'foo2': 'bar2'}]}
        dummy_headers = {'foo': 'bar', 'sample': 6}

        class DateBasedRestApiSourceWithHeaders(DateBasedRestAPISource):
            base_uri: Optional[str] = 'http://example.com'
            name: str = 'sample'

            def _get_headers(self) -> Dict[Any, Any]:
                return dummy_headers

            def _get_params(self, **kwargs) -> Dict[Any, Any]:
                return {}

        with patch.object(requests, 'get',
                          return_value=MockResponse(mock_data=mock_data)) as mock_get:
            source = DateBasedRestApiSourceWithHeaders(**self.options)
            data = list(source.get())

            url = f'{source.base_uri}/{source.name}'
            mock_get.assert_called_with(url=url, headers=dummy_headers, params={})

            self.assertListEqual(data, mock_data['results'])

    def test_get_with_default_headers(self):
        """Makes a GET request with the default headers passed during initialization"""
        mock_data = {'results': [{'foo': 'bar'}, {'foo2': 'bar2'}]}

        class DateBasedRestApiSourceWithDefaultHeaders(DateBasedRestAPISource):
            base_uri: Optional[str] = 'http://example.com'
            name: str = 'sample'

            def _get_params(self, **kwargs) -> Dict[Any, Any]:
                return {}

        with patch.object(requests, 'get',
                          return_value=MockResponse(mock_data=mock_data)) as mock_get:
            source = DateBasedRestApiSourceWithDefaultHeaders(**self.options)
            data = list(source.get())

            url = f'{source.base_uri}/{source.name}'
            mock_get.assert_called_with(url=url, headers=self.options['default_headers'], params={})

            self.assertListEqual(data, mock_data['results'])

    def test_get_with_authentication(self):
        """
        Makes a GET request with the appropriate authorization header,
        and attempts again if the authorization is expired
        """
        mock_data = {'results': [{'foo': 'bar'}, {'foo2': 'bar2'}]}

        def mocked_get_method(url, params, headers, **kwargs):
            if 'Authorization' in headers:
                return MockResponse(mock_data=mock_data)
            return MockResponse(status_code=401, ok=False)

        class DateBasedRestApiSourceWithAuthentication(DateBasedRestAPISource):
            base_uri: Optional[str] = 'http://example.com'
            name: str = 'sample'

            def _authenticate(self):
                self.default_headers['Authorization'] = 'Bearer some token'

            def _get_params(self, **kwargs) -> Dict[Any, Any]:
                return {}

        with patch.object(requests, 'get', side_effect=mocked_get_method) as mock_get:
            source = DateBasedRestApiSourceWithAuthentication(**self.options)
            data = list(source.get())

            uri = f'{source.base_uri}/{source.name}'
            calls = [call(url=uri, headers=self.options['default_headers'], params={}),
                     call(url=uri, headers={'Authorization': 'Bearer some token',
                                            **self.options['default_headers']}, params={})]
            mock_get.assert_has_calls(calls, any_order=False)

            self.assertListEqual(data, mock_data['results'])


if __name__ == '__main__':
    main()



"""Module containing base Destination Model class"""

from typing import Dict, Any, List


class DestinationBaseModel:

    @classmethod
    def get_attributes(cls) -> List[str]:
        """Returns the attributes/fields this model has"""
        raise NotImplementedError('get_attributes method not implemented.')

    @classmethod
    def get_last_saved_timestamp(cls):
        """Returns the last saved timestamp"""
        return None

    def update(self, *args, **kwargs):
        """
        Updates the attributes passed in the kwargs and saves
        """
        raise NotImplementedError('update method not implemented.')

    def save(self, *args, **kwargs):
        """Commits changes to the store"""
        raise NotImplementedError('save method not implemented.')

    def delete(self, *args, **kwargs):
        """Deletes the current instance"""
        raise NotImplementedError('delete method not implemented.')

    @classmethod
    def initialize(cls):
        """Initializes the model in the store e.g. creating tables in database"""
        raise NotImplementedError('initialize class method not implemented.')

    @classmethod
    def upsert(cls, data: Dict[Any, Any]) -> Dict[Any, Any]:
        """
        Updates the given instance or creates it if it does not exist
        Returns data so that it can be used by the next pipe
        """
        raise NotImplementedError('upsert method not implemented.')



"""
Database Connector class that provides a context manager for database connections
"""
import logging
from typing import Dict, Any, List, Optional

from sqlalchemy import create_engine, engine
from sqlalchemy.orm import scoped_session, sessionmaker, Session
from sqlalchemy.sql import text

from .config import DatabaseConnectionConfig


class DatabaseConnection:
    """
    Connects to a given database, returning a session in a context manager, and runs a given query
    """
    connections = {}

    def __init__(self, db_connection_config: DatabaseConnectionConfig):
        self.db_uri = db_connection_config.db_uri
        self.db_connection_config = db_connection_config
        self.db_session: Optional[scoped_session] = None

        if db_connection_config.db_uri.startswith('sqlite'):
            self.connection_engine: engine.Engine = create_engine(self.db_uri)
        else:
            self.connection_engine: engine.Engine = create_engine(
                self.db_uri,
                pool_size=self.db_connection_config.pool_size,
                max_overflow=self.db_connection_config.pool_size + 5,
                pool_timeout=self.db_connection_config.pool_timeout,
            )
        DatabaseConnection.connections[self.db_uri] = self

    def __enter__(self):
        """
        Creates a scoped session for the database that is thread-safe for reads
        """
        session = sessionmaker(
            bind=self.connection_engine,
            autocommit=self.db_connection_config.autocommit,
            autoflush=self.db_connection_config.autocommit,
            expire_on_commit=False,
        )
        self.db_session = scoped_session(session)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        This runs at the end of the 'with' statement as a way of cleaning up
        """
        if not self.db_connection_config.autocommit:
            # Attempt to commit any changes
            self.db_session.commit()

        self.db_session.remove()

    def close(self):
        """
        Closes the db connection engine
        """
        try:
            self.connection_engine.dispose()
            logging.info(f'Successfully closed connection {self.db_uri}')
        except Exception as exp:
            logging.warning(str(exp))

    def execute_sql(self, sql: str, params: Dict[str, Any] = {}) -> engine.ResultProxy:
        """
        Executes the sql passed as a parameter
        """
        return self.db_session.execute(text(sql), params=params)

    @classmethod
    def get_db_connection(cls, db_connection_config: DatabaseConnectionConfig):
        """
        Returns an instance of this class that has the given db_uri or if none, it creates a new one
        """
        db_connection = cls.connections.get(db_connection_config.db_uri, None)

        if db_connection is None:
            db_connection = cls(db_connection_config=db_connection_config)

        return db_connection

    @classmethod
    def close_all_connections(cls):
        """
        Closes all connections
        """
        for _, db_connection in cls.connections.items():
            db_connection.close()

    @classmethod
    def remove_connection(cls, db_uri: str):
        """
        Removes the given connection from the dictionary of connections
        """
        cls.connections.pop(db_uri, None)

    @classmethod
    def remove_all_connections(cls):
        """
        Removes all connection from the dictionary of connections
        """
        cls.connections.clear()

    @classmethod
    def open_connections(cls, db_configs: List[DatabaseConnectionConfig]):
        """
        Opens database connections if they are not open
        """
        for db_config in db_configs:
            cls.get_db_connection(db_connection_config=db_config)



"""Module containing a base controller class"""
import logging
import time
from datetime import datetime, timedelta
from typing import Type, Optional, Dict, Any, Iterator, List, Tuple

import pytz
from selenium.common.exceptions import WebDriverException
from urllib3.exceptions import ProtocolError

from ...destinations.base import DestinationBaseModel
from ...sources.base import BaseSource
from ...transformers.base import BaseTransformer
from email_notifier import ExceptionNotifier

from ...utils.cron import CronType
from ...utils.testing import TestingTimeoutError


class BaseController:
    """
    Class for the controller that gets data from a Source,
    transforms it
    and loads it in a destination basing on a given model
    """
    _source: BaseSource
    _interval: Optional[timedelta] = None
    _last_timestamp: Optional[datetime] = None

    destination_model_class: Type[DestinationBaseModel]
    source_class: Type[BaseSource]
    transformer_classes: List[Type[BaseTransformer]] = []
    interval_in_milliseconds: Optional[int] = None
    # cron_jobs override the interval; use one or the other
    cron_jobs: List[CronType] = []
    # exceptions
    _known_extraction_exceptions: Tuple[Exception] = (
        ProtocolError,
        WebDriverException,
        OSError,
        ConnectionError,
    )
    _known_transformation_exceptions: Tuple[Exception] = (OSError,)
    _known_loading_exceptions: Tuple[Exception] = (OSError,)
    _fatal_extraction_exceptions: Tuple[Exception] = (TestingTimeoutError,)
    _fatal_transformation_exceptions: Tuple[Exception] = (TestingTimeoutError,)
    _fatal_loading_exceptions: Tuple[Exception] = (TestingTimeoutError,)

    @classmethod
    def extract(cls) -> Iterator[Dict[Any, Any]]:
        """Initializes the extraction of the data from the source"""
        cls._initialize()
        # for intervals, no waiting before yield on first run.
        # for cron job, wait for the exact time before any yield
        should_wait_before_yielding = len(cls.cron_jobs) > 0

        while True:
            try:
                start_time = datetime.now(tz=pytz.utc)

                if should_wait_before_yielding:
                    cls._wait_for_interval_to_elapse(start_time_in_utc=start_time)

                yield from cls._query_source()

                if cls._interval is None and len(cls.cron_jobs) == 0:
                    break

                should_wait_before_yielding = True

            except cls._fatal_extraction_exceptions as exp:
                raise exp
            except cls._known_extraction_exceptions as exp:
                logging.error(f'{cls._source.name} Extraction \n{exp}')
                yield from []
            except Exception as unknown_exception:
                notifier = ExceptionNotifier(subject=f'[Unknown] {cls._source.name} Extraction')
                notifier.notify(unknown_exception)

    @classmethod
    def transform(cls, data: Optional[Dict[Any, Any]]) -> Iterator[Optional[Dict[Any, Any]]]:
        """Transforms the data row by row into something else"""
        if data is None:
            yield None

        try:
            transformed_data = data

            for transformer in cls.transformer_classes:
                if isinstance(transformed_data, list):
                    list_output = []

                    for datum in transformed_data:
                        data_from_current_transformer = transformer.run(data=datum)

                        # spread the returned list into the final output
                        if isinstance(data_from_current_transformer, list):
                            list_output = list_output + data_from_current_transformer
                        # append the returned item to the list output
                        else:
                            list_output.append(data_from_current_transformer)

                    transformed_data = list_output
                else:
                    transformed_data = transformer.run(data=transformed_data)

            if isinstance(transformed_data, list):
                yield from transformed_data
            else:
                yield transformed_data
        except cls._fatal_transformation_exceptions as exp:
            raise exp
        except cls._known_transformation_exceptions as exp:
            logging.error(f'{cls._source.name} Transformation \n{exp}')
            yield None
        except Exception as unknown_exception:
            notifier = ExceptionNotifier(subject=f'[Unknown] {cls._source.name} Transformation')
            notifier.notify(unknown_exception)

    @classmethod
    def load(cls, data: Optional[Dict[Any, Any]]) -> Iterator[Optional[Dict[Any, Any]]]:
        """Loads the data into the destination"""
        if data is None:
            yield None

        try:
            yield cls.destination_model_class.upsert(data=data)
        except cls._fatal_loading_exceptions as exp:
            raise exp
        except cls._known_loading_exceptions as exp:
            logging.error(f'{cls._source.name} Loading \n{exp}')
            yield None
        except Exception as unknown_exception:
            notifier = ExceptionNotifier(subject=f'[Unknown] {cls._source.name} Loading')
            notifier.notify(unknown_exception)

    @classmethod
    def _update_source(cls, *args, **kwargs) -> None:
        """
        Updates the configuration of the source especially
        those attributes that are bound to change
        """
        if not hasattr(cls, '_source'):
            cls._source = cls.source_class(
                attributes=cls.destination_model_class.get_attributes(),
                *args, **kwargs)

    @classmethod
    def _update_last_timestamp(cls):
        """Updates the last timestamp property of this class"""
        cls._last_timestamp = cls.destination_model_class.get_last_saved_timestamp()

    @classmethod
    def _update_interval(cls) -> None:
        """Updates the _interval basing on the interval_in_milliseconds"""
        if isinstance(cls.interval_in_milliseconds, int):
            cls._interval = timedelta(milliseconds=cls.interval_in_milliseconds)

    @classmethod
    def _initialize(cls) -> None:
        """Creates the tables if these are non-existent and updates the _interval if necessary"""
        cls._update_interval()
        cls.destination_model_class.initialize()
        cls._update_last_timestamp()

    @classmethod
    def _query_source(cls, ) -> Iterator[Dict[Any, Any]]:
        """Queries the data source"""
        cls._update_source()
        return cls._source.get()

    @classmethod
    def _wait_for_interval_to_elapse(cls, start_time_in_utc: datetime) -> None:
        """Waits for the interval to elapse before starting again"""
        next_timestamp = cls._get_closest_next_timestamp()

        if not isinstance(cls._interval, timedelta) and not isinstance(next_timestamp, datetime):
            return

        interval = cls._interval

        if next_timestamp is not None:
            interval = next_timestamp - start_time_in_utc

        time_used_up_so_far = datetime.now(tz=pytz.utc) - start_time_in_utc
        seconds_left_off_interval = (interval - time_used_up_so_far).total_seconds()

        if seconds_left_off_interval > 0:
            for index in range(round(seconds_left_off_interval)):
                time.sleep(1)

    @classmethod
    def _get_closest_next_timestamp(cls) -> Optional[datetime]:
        """
        Gets the timestamp from the available CronTypes that is the next to be run
        i.e. the earliest of all the available ‘next timestamps
        """
        timestamps = [cron_job.get_next_utc_timestamp() for cron_job in cls.cron_jobs]
        timestamps_without_nones = [timestamp for timestamp in timestamps if timestamp is not None]
        timestamps_without_nones.sort()

        try:
            return timestamps_without_nones[0]
        except IndexError:
            return None



"""Module containing the controller class for getting data from database to the database"""
from typing import Type

from ..base.date_based import DateBasedBaseController
from ...destinations.database.model import DatabaseBaseModel
from ...sources.database.date_based import DateBasedDatabaseSource


class DateBasedDatabaseToDatabaseController(DateBasedBaseController):
    """
    Class for the controller that gets data from the database
    and saves it in a database
    """
    _source: DateBasedDatabaseSource

    destination_model_class: Type[DatabaseBaseModel]
    source_class: Type[DateBasedDatabaseSource]



"""Module containing a data transformer that exchanges field names for their aliases"""
from typing import Dict, Any

from .base import BaseTransformer
from ..utils.data import replace_keys_with_aliases


class FieldNameAliasTransformer(BaseTransformer):
    """Changes the field names of the data passed to it to their aliases"""
    _alias_map: Dict[str, str] = {}

    @classmethod
    def run(cls, data: Dict[Any, Any]) -> Dict[Any, Any]:
        """Replaces the keys of the data to new ones"""
        return replace_keys_with_aliases(alias_map=cls._alias_map, data=data)



"""Module containing the controller class for getting data from the rest api to the database"""
from typing import Type

from ..base.date_based import DateBasedBaseController
from ...destinations.database.model import DatabaseBaseModel
from ...sources.rest_api.date_based import DateBasedRestAPISource


class DateBasedRestAPIToDatabaseController(DateBasedBaseController):
    """
    Class for the controller that gets data from the REST API
    and saves it in a database
    """
    _source: DateBasedRestAPISource

    destination_model_class: Type[DatabaseBaseModel]
    source_class: Type[DateBasedRestAPISource]



"""Tests for the field name cleaner transformer"""
from typing import Dict
from unittest import TestCase, main

from judah.transformers.field_name_cleaner_transformer import FieldNameCleanerTransformer


class TestFieldNameCleanerTransformer(TestCase):
    """Tests for the FieldNameCleanerTransformer base class"""

    def test_run(self):
        """
        should remove unwanted characters, replacing them with other characters in the field names
        """
        input_data = {
            '(nom)hey': 'John Doe',
            'address-district': 'Katakwi',
            'email': 'johndoe@example.com'
        }

        expected_output_data = {
            '_nom_hey': 'John Doe',
            'address district': 'Katakwi',
            'email': 'johndoe@example.com'
        }

        class ChildFieldNameCleanerTransformer(FieldNameCleanerTransformer):
            substring_replacement_map: Dict[str, str] = {
                '(': '_',
                ')': '_',
                '-': ' '
            }

        self.assertDictEqual(ChildFieldNameCleanerTransformer.run(input_data), expected_output_data)


if __name__ == '__main__':
    main()



"""Tests for the field name alias transformer"""
from typing import Dict
from unittest import TestCase, main

from judah.transformers.field_name_alias_transformer import FieldNameAliasTransformer


class TestFieldNameAliasTransformer(TestCase):
    """Tests for the FieldNameAliasTransformer base class"""

    def test_run(self):
        """
        should change the field names of the data passed to it to their aliases
        """
        input_data = {
            'nom': 'John Doe',
            'adresse': 'Katakwi',
            'email': 'johndoe@example.com'
        }

        expected_output_data = {
            'name': 'John Doe',
            'address': 'Katakwi',
            'email': 'johndoe@example.com'
        }

        class ChildFieldNameAliasTransformer(FieldNameAliasTransformer):
            _alias_map: Dict[str, str] = {
                'nom': 'name',
                'adresse': 'address'
            }

        self.assertDictEqual(ChildFieldNameAliasTransformer.run(input_data), expected_output_data)


if __name__ == '__main__':
    main()



from typing import Optional, Tuple, Iterator, Dict, Any

from ..base import BaseSource
from ...utils.dates import get_current_quarter, get_current_year, update_quarter_year_tuple, \
    get_default_start_quarter, get_default_start_year


class QuarterBasedBaseSource(BaseSource):
    start_quarter_and_year: Optional[Tuple[int, int]] = None
    end_quarter_and_year: Optional[Tuple[int, int]] = None
    default_batch_size_in_quarters: int = 1

    def get(self) -> Iterator[Dict[str, Any]]:
        """Returns data from a given date to a given date as an iterator"""
        start_quarter, start_year = self._get_next_start_quarter_and_year()
        end_quarter, end_year = self._get_next_end_quarter_and_year()

        if end_year < start_year:
            yield from []
        elif end_quarter < start_quarter and end_year <= start_year:
            yield from []
        else:
            yield from self._query_data_source(
                start_quarter_and_year=(start_quarter, start_year,),
                end_quarter_and_year=(end_quarter, end_year,))

            self.start_quarter_and_year = self._get_next_start_quarter_and_year(
                quarters_to_increment_by=self.default_batch_size_in_quarters)

    def _query_data_source(
            self, start_quarter_and_year: Tuple[int, int],
            end_quarter_and_year: Tuple[int, int]) -> Iterator[Dict[str, Any]]:
        """Queries a source given start and end quarter_year tuples and returns an iterator with data records"""
        raise NotImplementedError('Implement the _query_data_source method to return an iterator of dictionaries')

    def _get_next_end_quarter_and_year(self,
                                       quarters_to_increment_by: Optional[int] = 0,
                                       quarters_to_decrement_by: Optional[int] = 0) -> Tuple[int, int]:
        """Gets the next (end quarter, end year) tuple in the given iteration"""
        if isinstance(self.end_quarter_and_year, tuple):
            end_quarter, end_year = self.end_quarter_and_year
        elif isinstance(self.start_quarter_and_year, tuple):
            end_quarter, end_year = update_quarter_year_tuple(
                self.start_quarter_and_year,
                quarters_to_decrement_by=0,
                quarters_to_increment_by=self.default_batch_size_in_quarters)
        else:
            end_quarter, end_year = (get_current_quarter(), get_current_year(),)

        return update_quarter_year_tuple(quarter_year_tuple=(end_quarter, end_year,),
                                         quarters_to_increment_by=quarters_to_increment_by,
                                         quarters_to_decrement_by=quarters_to_decrement_by)

    def _get_next_start_quarter_and_year(self,
                                         quarters_to_increment_by: Optional[int] = 0,
                                         quarters_to_decrement_by: Optional[int] = 0) -> Tuple[int, int]:
        """Gets the next (start quarter, start year) tuple in the given iteration"""
        if isinstance(self.start_quarter_and_year, tuple):
            start_quarter, start_year = self.start_quarter_and_year
        elif isinstance(self.end_quarter_and_year, tuple):
            start_quarter, start_year = update_quarter_year_tuple(
                self.end_quarter_and_year,
                quarters_to_decrement_by=self.default_batch_size_in_quarters,
                quarters_to_increment_by=0)
        else:
            # default the end_quarter_and_year tuple to current quarter, year
            end_quarter_and_year_tuple = (get_current_quarter(), get_current_year())
            start_quarter, start_year = update_quarter_year_tuple(
                end_quarter_and_year_tuple,
                quarters_to_decrement_by=self.default_batch_size_in_quarters,
                quarters_to_increment_by=0)

        return update_quarter_year_tuple(quarter_year_tuple=(start_quarter, start_year,),
                                         quarters_to_increment_by=quarters_to_increment_by,
                                         quarters_to_decrement_by=quarters_to_decrement_by)
